"""
turbine_costsse_2015.py

Created by Janine Freeman 2015 based on turbine_costsse.py 2012.
Copyright (c) NREL. All rights reserved.
"""

from openmdao.api import ExplicitComponent, Problem, Group, IndepVarComp

import numpy as np

###### Rotor
#-------------------------------------------------------------------------------
class BladeCost2015(ExplicitComponent):

    def setup(self):


        # Inputs
        self.add_input('blade_mass',            0.0,  units='kg',     desc='component mass')
        self.add_input('blade_mass_cost_coeff', 14.6, units='USD/kg', desc='blade mass-cost coeff')
        self.add_input('blade_cost_external',   0.0,  units='USD',    desc='Blade cost computed by RotorSE')
        
        # Outputs
        self.add_output('blade_cost',           0.0,  units='USD',    desc='Overall wind turbine component capital costs excluding transportation costs')

    def compute(self, inputs, outputs):

        blade_mass = inputs['blade_mass']
        blade_mass_cost_coeff = inputs['blade_mass_cost_coeff']

        # calculate component cost
        if inputs['blade_cost_external'] < 1.:
            outputs['blade_cost'] = blade_mass_cost_coeff * blade_mass
        else:
            outputs['blade_cost'] = inputs['blade_cost_external']
        

# -----------------------------------------------------------------------------------------------
class HubCost2015(ExplicitComponent):

    def setup(self):


        # variables
        self.add_input('hub_mass', 0.0, desc='component mass', units='kg')
        self.add_input('hub_mass_cost_coeff', 3.9, desc='hub mass-cost coeff', units='USD/kg')
    
        # Outputs
        self.add_output('hub_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs):

        hub_mass_cost_coeff = inputs['hub_mass_cost_coeff']
        hub_mass = inputs['hub_mass']

        # calculate component cost
        HubCost2015 = hub_mass_cost_coeff * hub_mass
        outputs['hub_cost'] = HubCost2015
        

#-------------------------------------------------------------------------------
class PitchSystemCost2015(ExplicitComponent):

    def setup(self):


        # variables
        self.add_input('pitch_system_mass', 0.0, desc='component mass', units='kg')
        self.add_input('pitch_system_mass_cost_coeff', 22.1, desc='pitch system mass-cost coeff', units='USD/kg')
    
        # Outputs
        self.add_output('pitch_system_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs):
        
        pitch_system_mass = inputs['pitch_system_mass']
        pitch_system_mass_cost_coeff = inputs['pitch_system_mass_cost_coeff']
        
        #calculate system costs
        PitchSystemCost2015 = pitch_system_mass_cost_coeff * pitch_system_mass
        outputs['pitch_system_cost'] = PitchSystemCost2015
        
#-------------------------------------------------------------------------------
class SpinnerCost2015(ExplicitComponent):

    def setup(self):


        # variables
        self.add_input('spinner_mass', 0.0, desc='component mass', units='kg')
        self.add_input('spinner_mass_cost_coeff', 11.1, desc='spinner/nose cone mass-cost coeff', units='USD/kg')
    
        # Outputs
        self.add_output('spinner_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs):

        spinner_mass_cost_coeff = inputs['spinner_mass_cost_coeff']
        spinner_mass = inputs['spinner_mass']

        #calculate system costs
        SpinnerCost2015 = spinner_mass_cost_coeff * spinner_mass
        outputs['spinner_cost'] = SpinnerCost2015

#-------------------------------------------------------------------------------
class HubSystemCostAdder2015(ExplicitComponent):

    def setup(self):


        # Inputs
        self.add_input('hub_cost',          0.0, units='USD', desc='Hub component cost')
        self.add_input('hub_mass',          0.0, units='kg',  desc='Hub component mass')
        self.add_input('pitch_system_cost', 0.0, units='USD', desc='Pitch system cost')
        self.add_input('pitch_system_mass', 0.0, units='kg',  desc='Pitch system mass')
        self.add_input('spinner_cost',      0.0, units='USD', desc='Spinner component cost')
        self.add_input('spinner_mass',      0.0, units='kg', desc='Spinner component mass')
        self.add_input('hub_assemblyCostMultiplier',    0.0, desc='Rotor assembly cost multiplier')
        self.add_input('hub_overheadCostMultiplier',    0.0, desc='Rotor overhead cost multiplier')
        self.add_input('hub_profitMultiplier',          0.0, desc='Rotor profit multiplier')
        self.add_input('hub_transportMultiplier',       0.0, desc='Rotor transport multiplier')
    
        # Outputs
        self.add_output('hub_system_mass_tcc',  0.0, units='kg',  desc='Mass of the hub system, including hub, spinner, and pitch system for the blades')
        self.add_output('hub_system_cost',  0.0, units='USD', desc='Overall wind sub-assembly capial costs including transportation costs')

    def compute(self, inputs, outputs):

        hub_cost            = inputs['hub_cost']
        pitch_system_cost   = inputs['pitch_system_cost']
        spinner_cost        = inputs['spinner_cost']
        
        hub_mass            = inputs['hub_mass']
        pitch_system_mass   = inputs['pitch_system_mass']
        spinner_mass        = inputs['spinner_mass']
        
        hub_assemblyCostMultiplier  = inputs['hub_assemblyCostMultiplier']
        hub_overheadCostMultiplier  = inputs['hub_overheadCostMultiplier']
        hub_profitMultiplier        = inputs['hub_profitMultiplier']
        hub_transportMultiplier     = inputs['hub_transportMultiplier']

        # Updated calculations below to account for assembly, transport, overhead and profit
        outputs['hub_system_mass_tcc'] = hub_mass + pitch_system_mass + spinner_mass
        partsCost = hub_cost + pitch_system_cost + spinner_cost
        outputs['hub_system_cost'] = (1 + hub_transportMultiplier + hub_profitMultiplier) * ((1 + hub_overheadCostMultiplier + hub_assemblyCostMultiplier) * partsCost)

#-------------------------------------------------------------------------------
class RotorCostAdder2015(ExplicitComponent):
    """
    RotorCostAdder adds up individual rotor system and component costs to get overall rotor cost.
    """

    def setup(self):
        

        # Inputs
        self.add_input('blade_cost',        0.0, units='USD',   desc='Individual blade cost')
        self.add_input('blade_mass',        0.0, units='kg',    desc='Individual blade mass')
        self.add_input('hub_system_cost',   0.0, units='USD',   desc='Cost for hub system')
        self.add_input('hub_system_mass_tcc',   0.0, units='kg',    desc='Mass for hub system')
        self.add_discrete_input('blade_number',      3,                  desc='Number of rotor blades')
    
        # Outputs
        self.add_output('rotor_cost',       0.0, units='USD',   desc='Overall wind sub-assembly capial costs including transportation costs')
        self.add_output('rotor_mass_tcc',   0.0, units='kg',    desc='Rotor mass, including blades, pitch system, hub, and spinner')
        
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        blade_cost      = inputs['blade_cost']
        blade_mass      = inputs['blade_mass']
        blade_number    = discrete_inputs['blade_number']
        hub_system_cost = inputs['hub_system_cost']
        hub_system_mass = inputs['hub_system_mass_tcc']

        outputs['rotor_cost']      = blade_cost * blade_number + hub_system_cost
        outputs['rotor_mass_tcc']  = blade_mass * blade_number + hub_system_mass

#-------------------------------------------------------------------------------


###### Nacelle
# -------------------------------------------------
class LowSpeedShaftCost2015(ExplicitComponent):

    def setup(self):


        # variables
        self.add_input('lss_mass', 0.0, desc='component mass', units='kg') #mass input
        self.add_input('lss_mass_cost_coeff', 11.9, desc='low speed shaft mass-cost coeff', units='USD/kg')
    
        # Outputs
        self.add_output('lss_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs') #initialize cost output

    def compute(self, inputs, outputs):

        lss_mass_cost_coeff = inputs['lss_mass_cost_coeff']
        lss_mass = inputs['lss_mass']

        # calculate component cost
        outputs['lss_cost'] = lss_mass_cost_coeff * lss_mass

#-------------------------------------------------------------------------------
class BearingsCost2015(ExplicitComponent):

    def setup(self):


        # variables
        self.add_input('main_bearing_mass', 0.0, desc='component mass', units='kg') #mass input
        self.add_discrete_input('main_bearing_number', 2, desc='number of main bearings') #number of main bearings- defaults to 2
        self.add_input('bearings_mass_cost_coeff', 4.5, desc='main bearings mass-cost coeff', units='USD/kg')
    
        # Outputs
        self.add_output('main_bearing_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        main_bearing_mass = inputs['main_bearing_mass']
        main_bearing_number = discrete_inputs['main_bearing_number']
        bearings_mass_cost_coeff = inputs['bearings_mass_cost_coeff']

        #calculate component cost 
        outputs['main_bearing_cost'] = bearings_mass_cost_coeff * main_bearing_mass * main_bearing_number

#-------------------------------------------------------------------------------
class GearboxCost2015(ExplicitComponent):

    def setup(self):


        # variables
        self.add_input('gearbox_mass', 0.0, units='kg', desc='component mass')
        self.add_input('gearbox_mass_cost_coeff', 12.9, desc='gearbox mass-cost coeff', units='USD/kg')
    
        # Outputs
        self.add_output('gearbox_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs):

        gearbox_mass = inputs['gearbox_mass']
        gearbox_mass_cost_coeff = inputs['gearbox_mass_cost_coeff']

        outputs['gearbox_cost'] = gearbox_mass_cost_coeff * gearbox_mass

#-------------------------------------------------------------------------------
class HighSpeedSideCost2015(ExplicitComponent):

    def setup(self):
        

        # variables
        self.add_input('hss_mass', 0.0, desc='component mass', units='kg')
        self.add_input('hss_mass_cost_coeff', 6.8, desc='high speed side mass-cost coeff', units='USD/kg')
    
        # Outputs
        self.add_output('hss_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs):

        hss_mass = inputs['hss_mass']
        hss_mass_cost_coeff = inputs['hss_mass_cost_coeff']
        
        outputs['hss_cost'] = hss_mass_cost_coeff * hss_mass

#-------------------------------------------------------------------------------
class GeneratorCost2015(ExplicitComponent):

    def setup(self):


        # variables
        self.add_input('generator_mass', 0.0, desc='component mass', units='kg')
        self.add_input('generator_mass_cost_coeff', 12.4, desc='generator mass cost coeff', units='USD/kg')
    
        # Outputs
        self.add_output('generator_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs):

        generator_mass = inputs['generator_mass']
        generator_mass_cost_coeff = inputs['generator_mass_cost_coeff']
        
        outputs['generator_cost'] = generator_mass_cost_coeff * generator_mass

#-------------------------------------------------------------------------------
class BedplateCost2015(ExplicitComponent):

    def setup(self):
        
        
        # variables
        self.add_input('bedplate_mass', 0.0, desc='component mass', units='kg')
        self.add_input('bedplate_mass_cost_coeff', 2.9, desc='bedplate mass-cost coeff', units='USD/kg')
    
        # Outputs
        self.add_output('bedplate_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs):

        bedplate_mass = inputs['bedplate_mass']
        bedplate_mass_cost_coeff = inputs['bedplate_mass_cost_coeff']

        outputs['bedplate_cost'] = bedplate_mass_cost_coeff * bedplate_mass

#---------------------------------------------------------------------------------
class YawSystemCost2015(ExplicitComponent):

    def setup(self):


        # variables
        self.add_input('yaw_mass', 0.0, desc='component mass', units='kg')
        self.add_input('yaw_mass_cost_coeff', 8.3, desc='yaw system mass cost coeff', units='USD/kg')
    
        # Outputs
        self.add_output('yaw_system_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs):

        yaw_mass = inputs['yaw_mass']
        yaw_mass_cost_coeff = inputs['yaw_mass_cost_coeff']
        
        outputs['yaw_system_cost'] = yaw_mass_cost_coeff * yaw_mass

#---------------------------------------------------------------------------------
class VariableSpeedElecCost2015(ExplicitComponent):

    def setup(self):


        # variables
        self.add_input('vs_electronics_mass', 0.0, desc='component mass', units='kg')
        self.add_input('vs_electronics_mass_cost_coeff', 18.8, desc='variable speed electronics mass cost coeff', units='USD/kg')
    
        # Outputs
        self.add_output('vs_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs):

        vs_electronics_mass = inputs['vs_electronics_mass']
        vs_electronics_mass_cost_coeff = inputs['vs_electronics_mass_cost_coeff']

        outputs['vs_cost'] = vs_electronics_mass_cost_coeff * vs_electronics_mass

#---------------------------------------------------------------------------------
class HydraulicCoolingCost2015(ExplicitComponent):

    def setup(self):


        # variables
        self.add_input('hvac_mass', 0.0, desc='component mass', units='kg')
        self.add_input('hvac_mass_cost_coeff', 124.0, desc='hydraulic and cooling system mass cost coeff', units='USD/kg')
    
        # Outputs
        self.add_output('hvac_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs):

        hvac_mass = inputs['hvac_mass']
        hvac_mass_cost_coeff = inputs['hvac_mass_cost_coeff']

        # calculate cost
        outputs['hvac_cost'] = hvac_mass_cost_coeff * hvac_mass

#---------------------------------------------------------------------------------
class NacelleCoverCost2015(ExplicitComponent):

    def setup(self):


        # variables
        self.add_input('cover_mass', 0.0, desc='component mass', units='kg')
        self.add_input('cover_mass_cost_coeff', 5.7, desc='nacelle cover mass cost coeff', units='USD/kg')
    
        # Outputs
        self.add_output('cover_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs):

        cover_mass = inputs['cover_mass']
        cover_mass_cost_coeff = inputs['cover_mass_cost_coeff']

        outputs['cover_cost'] = cover_mass_cost_coeff * cover_mass

#---------------------------------------------------------------------------------
class ElecConnecCost2015(ExplicitComponent):

    def setup(self):


        # variables
        self.add_input('machine_rating', 0.0, desc='machine rating', units='kW')
        self.add_input('elec_connec_machine_rating_cost_coeff', 41.85, units='USD/kW', desc='electrical connections cost coefficient per kW')
    
        # Outputs
        self.add_output('elec_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs):

        machine_rating = inputs['machine_rating']
        elec_connec_machine_rating_cost_coeff = inputs['elec_connec_machine_rating_cost_coeff']

        outputs['elec_cost'] = elec_connec_machine_rating_cost_coeff * machine_rating


#---------------------------------------------------------------------------------
class ControlsCost2015(ExplicitComponent):

    def setup(self):


        # variables
        self.add_input('machine_rating', 0.0, desc='machine rating', units='kW')
        self.add_input('controls_machine_rating_cost_coeff', 21.15, units='USD/kW', desc='controls cost coefficient per kW')
    
        # Outputs
        self.add_output('controls_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs):

        machine_rating = inputs['machine_rating']
        coeff          = inputs['controls_machine_rating_cost_coeff']

        outputs['controls_cost'] = machine_rating * coeff

#---------------------------------------------------------------------------------
class OtherMainframeCost2015(ExplicitComponent):

    def setup(self):  


        # variables
        self.add_input('platforms_mass', 0.0, desc='component mass', units='kg')
        self.add_input('platforms_mass_cost_coeff', 17.1, desc='nacelle platforms mass cost coeff', units='USD/kg')
        self.add_discrete_input('crane', False, desc='flag for presence of onboard crane')
        self.add_input('crane_cost', 12000.0, desc='crane cost if present', units='USD')
        # self.add_input('bedplate_cost', 0.0, desc='component cost', units='USD')
        # self.add_input('base_hardware_cost_coeff', 0.7, desc='base hardware cost coeff based on bedplate cost')
    
        # Outputs
        self.add_output('other_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        platforms_mass = inputs['platforms_mass']
        platforms_mass_cost_coeff = inputs['platforms_mass_cost_coeff']
        crane = discrete_inputs['crane']
        crane_cost = inputs['crane_cost']
        # bedplate_cost = inputs['bedplate_cost']
        # base_hardware_cost_coeff = inputs['base_hardware_cost_coeff']

        # nacelle platform cost

        # crane cost
        if (crane):
            craneCost  = crane_cost
            craneMass  = 3e3
            NacellePlatformsCost = platforms_mass_cost_coeff * (platforms_mass - craneMass)
        else:
            craneCost  = 0.0
            NacellePlatformsCost = platforms_mass_cost_coeff * platforms_mass

        # base hardware cost
        #BaseHardwareCost = bedplate_cost * base_hardware_cost_coeff
    
        #aggregate all three mainframe costs
        outputs['other_cost'] = NacellePlatformsCost + craneCost #+ BaseHardwareCost

#-------------------------------------------------------------------------------
class TransformerCost2015(ExplicitComponent):

    def setup(self):


        # variables
        self.add_input('transformer_mass', 0.0, desc='component mass', units='kg')
        self.add_input('transformer_mass_cost_coeff', 18.8, desc='transformer mass cost coeff', units='USD/kg') #mass-cost coeff with default from ppt
    
        # Outputs
        self.add_output('transformer_cost', 0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs):

        transformer_mass = inputs['transformer_mass']
        transformer_mass_cost_coeff = inputs['transformer_mass_cost_coeff']
        
        outputs['transformer_cost'] = transformer_mass_cost_coeff * transformer_mass

#-------------------------------------------------------------------------------
class NacelleSystemCostAdder2015(ExplicitComponent):

    def setup(self):
        

        # variables
        self.add_input('lss_cost',          0.0, units='USD', desc='Component cost')
        self.add_input('lss_mass',          0.0, units='kg',  desc='Component mass')
        self.add_input('main_bearing_cost', 0.0, units='USD', desc='Component cost')
        self.add_input('main_bearing_mass', 0.0, units='kg',  desc='Component mass')
        self.add_input('gearbox_cost',      0.0, units='USD', desc='Component cost')
        self.add_input('gearbox_mass',      0.0, units='kg',  desc='Component mass')
        self.add_input('hss_cost',          0.0, units='USD', desc='Component cost')
        self.add_input('hss_mass',          0.0, units='kg',  desc='Component mass')
        self.add_input('generator_cost',    0.0, units='USD', desc='Component cost')
        self.add_input('generator_mass',    0.0, units='kg',  desc='Component mass')
        self.add_input('bedplate_cost',     0.0, units='USD', desc='Component cost')
        self.add_input('bedplate_mass',     0.0, units='kg',  desc='Component mass')
        self.add_input('yaw_system_cost',   0.0, units='USD', desc='Component cost')
        self.add_input('yaw_mass',          0.0, units='kg',  desc='Component mass')
        self.add_input('vs_cost',           0.0, units='USD', desc='Component cost')
        self.add_input('vs_mass',           0.0, units='kg',  desc='Component mass')
        self.add_input('hvac_cost',         0.0, units='USD', desc='Component cost')
        self.add_input('hvac_mass',         0.0, units='kg',  desc='Component mass')
        self.add_input('cover_cost',        0.0, units='USD', desc='Component cost')
        self.add_input('cover_mass',        0.0, units='kg',  desc='Component mass')
        self.add_input('elec_cost',         0.0, units='USD', desc='Component cost')
        self.add_input('controls_cost',     0.0, units='USD', desc='Component cost')
        self.add_input('other_cost',        0.0, units='USD', desc='Component cost')
        self.add_input('transformer_cost',  0.0, units='USD', desc='Component cost')
        self.add_input('transformer_mass',  0.0, units='kg',  desc='Component mass')
        self.add_discrete_input('main_bearing_number', 2, desc ='number of bearings')
        
        #multipliers
        self.add_input('nacelle_assemblyCostMultiplier', 0.0, desc='nacelle assembly cost multiplier')
        self.add_input('nacelle_overheadCostMultiplier', 0.0, desc='nacelle overhead cost multiplier')
        self.add_input('nacelle_profitMultiplier',       0.0, desc='nacelle profit multiplier')
        self.add_input('nacelle_transportMultiplier',    0.0, desc='nacelle transport multiplier')
    
        # returns
        self.add_output('nacelle_cost', 0.0, units='USD', desc='component cost')
        self.add_output('nacelle_mass_tcc', 0.0, units='kg',  desc='Nacelle mass, with all nacelle components, without the rotor')

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        lss_cost            = inputs['lss_cost']
        main_bearing_cost   = inputs['main_bearing_cost']
        gearbox_cost        = inputs['gearbox_cost']
        hss_cost            = inputs['hss_cost']
        generator_cost      = inputs['generator_cost']
        bedplate_cost       = inputs['bedplate_cost']
        yaw_system_cost     = inputs['yaw_system_cost']
        vs_cost             = inputs['vs_cost']
        hvac_cost           = inputs['hvac_cost']
        cover_cost          = inputs['cover_cost']
        elec_cost           = inputs['elec_cost']
        controls_cost       = inputs['controls_cost']
        other_cost          = inputs['other_cost']
        transformer_cost    = inputs['transformer_cost']
        
        lss_mass            = inputs['lss_mass']
        main_bearing_mass   = inputs['main_bearing_mass']
        gearbox_mass        = inputs['gearbox_mass']
        hss_mass            = inputs['hss_mass']
        generator_mass      = inputs['generator_mass']
        bedplate_mass       = inputs['bedplate_mass']
        yaw_mass            = inputs['yaw_mass']
        vs_mass             = inputs['vs_mass']
        hvac_mass           = inputs['hvac_mass']
        cover_mass          = inputs['cover_mass']
        transformer_mass    = inputs['transformer_mass']
        
        main_bearing_number = discrete_inputs['main_bearing_number']

        nacelle_assemblyCostMultiplier  = inputs['nacelle_assemblyCostMultiplier']
        nacelle_overheadCostMultiplier  = inputs['nacelle_overheadCostMultiplier']
        nacelle_profitMultiplier        = inputs['nacelle_profitMultiplier']
        nacelle_transportMultiplier     = inputs['nacelle_transportMultiplier']        

        #apply multipliers for assembly, transport, overhead, and profits
        outputs['nacelle_mass_tcc'] = lss_mass + main_bearing_number * main_bearing_mass + gearbox_mass + hss_mass + generator_mass + bedplate_mass + yaw_mass + vs_mass + hvac_mass + cover_mass + transformer_mass
        partsCost = lss_cost + main_bearing_number * main_bearing_cost + gearbox_cost + hss_cost + generator_cost + bedplate_cost + yaw_system_cost + vs_cost + hvac_cost + cover_cost + elec_cost + controls_cost + other_cost + transformer_cost
        outputs['nacelle_cost'] = (1 + nacelle_transportMultiplier + nacelle_profitMultiplier) * ((1 + nacelle_overheadCostMultiplier + nacelle_assemblyCostMultiplier) * partsCost)

###### Tower
#-------------------------------------------------------------------------------
class TowerCost2015(ExplicitComponent):

    def setup(self):
      

        # variables
        self.add_input('tower_mass',            0.0, units='kg',     desc='tower mass')
        self.add_input('tower_mass_cost_coeff', 2.9, units='USD/kg', desc='tower mass-cost coeff') #mass-cost coeff with default from ppt
        self.add_input('tower_cost_external',   0.0, units='USD',    desc='Tower cost computed by TowerSE')
        
        # Outputs
        self.add_output('tower_parts_cost',     0.0, units='USD', desc='Overall wind turbine component capial costs excluding transportation costs')

    def compute(self, inputs, outputs):

        tower_mass = inputs['tower_mass']
        tower_mass_cost_coeff = inputs['tower_mass_cost_coeff']
        
        # calculate component cost
        if inputs['tower_cost_external'] < 1.:
            outputs['tower_parts_cost'] = tower_mass_cost_coeff * tower_mass
        else:
            outputs['tower_parts_cost'] = inputs['tower_cost_external']
        
        
#-------------------------------------------------------------------------------
class TowerCostAdder2015(ExplicitComponent):

    def setup(self):


        # variables
        self.add_input('tower_parts_cost', 0.0, units='USD', desc='component cost')
      
        # multipliers
        self.add_input('tower_assemblyCostMultiplier', 0.0, desc='tower assembly cost multiplier')
        self.add_input('tower_overheadCostMultiplier', 0.0, desc='tower overhead cost multiplier')
        self.add_input('tower_profitMultiplier', 0.0, desc='tower profit cost multiplier')
        self.add_input('tower_transportMultiplier', 0.0, desc='tower transport cost multiplier')
        
        # returns
        self.add_output('tower_cost', 0.0, units='USD', desc='component cost') 

    def compute(self, inputs, outputs):

        tower_parts_cost = inputs['tower_parts_cost']

        tower_assemblyCostMultiplier = inputs['tower_assemblyCostMultiplier']
        tower_overheadCostMultiplier = inputs['tower_overheadCostMultiplier']
        tower_profitMultiplier = inputs['tower_profitMultiplier']
        tower_transportMultiplier = inputs['tower_transportMultiplier']

        partsCost = tower_parts_cost
        outputs['tower_cost'] = (1 + tower_transportMultiplier + tower_profitMultiplier) * ((1 + tower_overheadCostMultiplier + tower_assemblyCostMultiplier) * partsCost)

#-------------------------------------------------------------------------------
class TurbineCostAdder2015(ExplicitComponent):

    def setup(self):


        # Variables
        self.add_input('rotor_cost',        0.0, units='USD',   desc='Rotor cost')
        self.add_input('rotor_mass_tcc',    0.0, units='kg',    desc='Rotor mass')
        self.add_input('nacelle_cost',      0.0, units='USD',   desc='Nacelle cost')
        self.add_input('nacelle_mass_tcc',      0.0, units='kg',    desc='Nacelle mass')
        self.add_input('tower_cost',        0.0, units='USD',   desc='Tower cost')
        self.add_input('tower_mass',        0.0, units='kg',    desc='Tower mass')
        self.add_input('machine_rating',    0.0, units='kW',    desc='Machine rating')
    
        # parameters
        self.add_input('turbine_assemblyCostMultiplier',    0.0, desc='Turbine multiplier for assembly cost in manufacturing')
        self.add_input('turbine_overheadCostMultiplier',    0.0, desc='Turbine multiplier for overhead')
        self.add_input('turbine_profitMultiplier',          0.0, desc='Turbine multiplier for profit markup')
        self.add_input('turbine_transportMultiplier',       0.0, desc='Turbine multiplier for transport costs')
    
        # Outputs
        self.add_output('turbine_mass_tcc',     0.0, units='kg',    desc='Turbine total mass, without foundation')
        self.add_output('turbine_cost',     0.0, units='USD',   desc='Overall wind turbine capital costs including transportation costs')
        self.add_output('turbine_cost_kW',  0.0, units='USD/kW',desc='Overall wind turbine capial costs including transportation costs')
        
    def compute(self, inputs, outputs):

        rotor_cost      = inputs['rotor_cost']
        nacelle_cost    = inputs['nacelle_cost']
        tower_cost      = inputs['tower_cost']
        
        rotor_mass_tcc  = inputs['rotor_mass_tcc']
        nacelle_mass_tcc= inputs['nacelle_mass_tcc']
        tower_mass      = inputs['tower_mass']
        
        turbine_assemblyCostMultiplier = inputs['turbine_assemblyCostMultiplier']
        turbine_overheadCostMultiplier = inputs['turbine_overheadCostMultiplier']
        turbine_profitMultiplier = inputs['turbine_profitMultiplier']
        turbine_transportMultiplier = inputs['turbine_transportMultiplier']

        partsCost = rotor_cost + nacelle_cost + tower_cost
        
        
        outputs['turbine_mass_tcc']    =  rotor_mass_tcc + nacelle_mass_tcc + tower_mass
        outputs['turbine_cost']    = (1 + turbine_transportMultiplier + turbine_profitMultiplier) * ((1 + turbine_overheadCostMultiplier + turbine_assemblyCostMultiplier) * partsCost)
        outputs['turbine_cost_kW'] = outputs['turbine_cost'] / inputs['machine_rating']

class Outputs2Screen(ExplicitComponent):
    def initialize(self):
        self.options.declare('verbosity', default=False)
        
    def setup(self):
        
        self.add_input('blade_cost',       0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('blade_mass',       0.0,  units='kg',  desc='Blade mass')
        self.add_input('hub_cost',         0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('hub_mass',         0.0,  units='kg',  desc='Hub mass')
        self.add_input('pitch_system_cost',       0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('pitch_system_mass',0.0,  units='kg',  desc='Pitch system mass')
        self.add_input('spinner_cost',     0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('spinner_mass',     0.0,  units='kg',  desc='Spinner mass')
        self.add_input('lss_cost',         0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('lss_mass',         0.0,  units='kg',  desc='LSS mass')
        self.add_input('main_bearing_cost',0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('main_bearing_mass',0.0,  units='kg',  desc='Main bearing mass')
        self.add_input('gearbox_cost',     0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('gearbox_mass',     0.0,  units='kg',  desc='LSS mass')
        self.add_input('hss_cost',         0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('hss_mass',         0.0,  units='kg',  desc='HSS mass')
        self.add_input('generator_cost',   0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('generator_mass',   0.0,  units='kg',  desc='Generator mass')
        self.add_input('bedplate_cost',    0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('bedplate_mass',    0.0,  units='kg',  desc='Bedplate mass')
        self.add_input('yaw_system_cost',  0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('yaw_mass',         0.0,  units='kg',  desc='Yaw system mass')
        self.add_input('hvac_cost',        0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('hvac_mass',        0.0,  units='kg',  desc='HVAC mass')
        self.add_input('cover_cost',       0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('cover_mass',       0.0,  units='kg',  desc='Cover mass')
        self.add_input('elec_cost',        0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('controls_cost',    0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('other_cost',       0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('transformer_cost', 0.0,  units='USD', desc='Overall wind turbine component capital costs excluding transportation costs')
        self.add_input('transformer_mass', 0.0,  units='kg',  desc='Transformer mass')
                                           
        self.add_input('rotor_cost',       0.0,  units='USD', desc='Overall wind turbine rotor capital costs')
        self.add_input('rotor_mass_tcc',   0.0,  units='kg',  desc='Rotor mass')
        self.add_input('nacelle_cost',     0.0,  units='USD', desc='Overall wind turbine nacelle capital costs')
        self.add_input('nacelle_mass_tcc', 0.0,  units='kg',  desc='Nacelle mass')
        self.add_input('tower_cost',       0.0,  units='USD', desc='Overall wind turbine tower capital costs')
        self.add_input('tower_mass',       0.0,  units='kg',  desc='Tower mass')
        self.add_input('turbine_cost',     0.0,  units='USD', desc='Overall wind turbine capital costs including transportation costs')
        self.add_input('turbine_cost_kW',  0.0,  units='USD/kW', desc='Overall wind turbine capital costs including transportation costs per kW')
        self.add_input('turbine_mass_tcc', 0.0,  units='kg',  desc='Turbine mass')
        
        
    def compute(self, inputs, outputs):        
        
        if self.options['verbosity'] == True:
        
            
            print('################################################')
            print('Computation of costs of the main turbine components from TurbineCostSE')
            print('Blade cost              %.3f k USD       mass %.3f kg' % (inputs['blade_cost'] * 1.e-003,        inputs['blade_mass']))
            print('Pitch system cost       %.3f k USD       mass %.3f kg' % (inputs['pitch_system_cost'] * 1.e-003, inputs['pitch_system_mass']))
            print('Hub cost                %.3f k USD       mass %.3f kg' % (inputs['hub_cost'] * 1.e-003,          inputs['hub_mass']))
            print('Spinner cost            %.3f k USD       mass %.3f kg' % (inputs['spinner_cost'] * 1.e-003,      inputs['spinner_mass']))
            print('------------------------------------------------')
            print('Rotor cost              %.3f k USD       mass %.3f kg' % (inputs['rotor_cost'] * 1.e-003,        inputs['rotor_mass_tcc']))
            print('')
            print('LSS cost                %.3f k USD       mass %.3f kg' % (inputs['lss_cost'] * 1.e-003,          inputs['lss_mass']))
            print('Main bearing cost       %.3f k USD       mass %.3f kg' % (inputs['main_bearing_cost'] * 1.e-003, inputs['main_bearing_mass']))
            print('Gearbox cost            %.3f k USD       mass %.3f kg' % (inputs['gearbox_cost'] * 1.e-003,      inputs['gearbox_mass']))
            print('HSS cost                %.3f k USD       mass %.3f kg' % (inputs['hss_cost'] * 1.e-003,          inputs['hss_mass']))
            print('Generator cost          %.3f k USD       mass %.3f kg' % (inputs['generator_cost'] * 1.e-003,    inputs['generator_mass']))
            print('Bedplate cost           %.3f k USD       mass %.3f kg' % (inputs['bedplate_cost'] * 1.e-003,     inputs['bedplate_mass']))
            print('Yaw system cost         %.3f k USD       mass %.3f kg' % (inputs['yaw_system_cost'] * 1.e-003,   inputs['yaw_mass']))
            print('HVAC cost               %.3f k USD       mass %.3f kg' % (inputs['hvac_cost'] * 1.e-003,         inputs['hvac_mass']))
            print('Nacelle cover cost      %.3f k USD       mass %.3f kg' % (inputs['cover_cost'] * 1.e-003,        inputs['cover_mass']))
            print('Electr connection cost  %.3f k USD'                    % (inputs['elec_cost'] * 1.e-003))
            print('Controls cost           %.3f k USD'                    % (inputs['controls_cost'] * 1.e-003))
            print('Other main frame cost   %.3f k USD'                    % (inputs['other_cost'] * 1.e-003))
            print('Transformer cost        %.3f k USD       mass %.3f kg' % (inputs['transformer_cost'] * 1.e-003,  inputs['transformer_mass']))
            print('------------------------------------------------')
            print('Nacelle cost            %.3f k USD       mass %.3f kg' % (inputs['nacelle_cost'] * 1.e-003,      inputs['nacelle_mass_tcc']))
            print('')
            print('Tower cost              %.3f k USD       mass %.3f kg' % (inputs['tower_cost'] * 1.e-003,        inputs['tower_mass']))
            print('------------------------------------------------')
            print('------------------------------------------------')
            print('Turbine cost            %.3f k USD       mass %.3f kg' % (inputs['turbine_cost'] * 1.e-003,      inputs['turbine_mass_tcc']))
            print('Turbine cost per kW     %.3f k USD/kW'                 % inputs['turbine_cost_kW'])
            print('################################################')
                
    

#-------------------------------------------------------------------------------
class Turbine_CostsSE_2015(Group):
    def initialize(self):
        self.options.declare('verbosity', default=False)
        self.options.declare('topLevelFlag', default=True)

    def setup(self):
        self.verbosity = self.options['verbosity']

        if self.options['topLevelFlag']:
            sharedIndeps = IndepVarComp()
            sharedIndeps.add_output('machine_rating',     units='kW', val=0.0)
            sharedIndeps.add_output('blade_mass',         units='kg', val=0.0)
            sharedIndeps.add_output('hub_mass',           units='kg', val=0.0)
            sharedIndeps.add_output('pitch_system_mass',  units='kg', val=0.0)
            sharedIndeps.add_output('spinner_mass',       units='kg', val=0.0)
            sharedIndeps.add_output('lss_mass',           units='kg', val=0.0)
            sharedIndeps.add_output('bearings_mass',      units='kg', val=0.0)
            sharedIndeps.add_output('gearbox_mass',       units='kg', val=0.0)
            sharedIndeps.add_output('main_bearing_mass',  units='kg', val=0.0)
            sharedIndeps.add_discrete_output('main_bearing_number',  val=0)
            sharedIndeps.add_output('hss_mass',           units='kg', val=0.0)
            sharedIndeps.add_output('generator_mass',     units='kg', val=0.0)
            sharedIndeps.add_output('bedplate_mass',      units='kg', val=0.0)
            sharedIndeps.add_output('yaw_mass',           units='kg', val=0.0)
            sharedIndeps.add_output('vs_electronics_mass',units='kg', val=0.0)
            sharedIndeps.add_output('hvac_mass',          units='kg', val=0.0)
            sharedIndeps.add_output('cover_mass',         units='kg', val=0.0)
            sharedIndeps.add_output('platforms_mass',     units='kg', val=0.0)
            sharedIndeps.add_output('transformer_mass',   units='kg', val=0.0)
            sharedIndeps.add_output('tower_mass',         units='kg', val=0.0)
            sharedIndeps.add_discrete_output('crane', val=False)
            self.add_subsystem('sharedIndeps', sharedIndeps, promotes=['*'])
        
        indeps = IndepVarComp()
        indeps.add_output('blade_mass_cost_coeff',         units='USD/kg', val=14.6)
        indeps.add_output('hub_mass_cost_coeff',           units='USD/kg', val=3.9)
        indeps.add_output('pitch_system_mass_cost_coeff',  units='USD/kg', val=22.1)
        indeps.add_output('spinner_mass_cost_coeff',       units='USD/kg', val=11.1)
        indeps.add_output('lss_mass_cost_coeff',           units='USD/kg', val=11.9)
        indeps.add_output('bearings_mass_cost_coeff',      units='USD/kg', val=4.5)
        indeps.add_output('gearbox_mass_cost_coeff',       units='USD/kg', val=12.9)
        indeps.add_output('hss_mass_cost_coeff',           units='USD/kg', val=6.8)
        indeps.add_output('generator_mass_cost_coeff',     units='USD/kg', val=12.4)
        indeps.add_output('bedplate_mass_cost_coeff',      units='USD/kg', val=2.9)
        indeps.add_output('yaw_mass_cost_coeff',           units='USD/kg', val=8.3)
        indeps.add_output('vs_electronics_mass_cost_coeff',units='USD/kg', val=18.8)
        indeps.add_output('hvac_mass_cost_coeff',          units='USD/kg', val=124.0)
        indeps.add_output('cover_mass_cost_coeff',         units='USD/kg', val=5.7)
        indeps.add_output('elec_connec_machine_rating_cost_coeff',units='USD/kW', val=41.85)
        indeps.add_output('platforms_mass_cost_coeff',     units='USD/kg', val=17.1)
        indeps.add_output('base_hardware_cost_coeff',      units='USD/kg', val=0.7)
        indeps.add_output('transformer_mass_cost_coeff',   units='USD/kg', val=18.8)
        indeps.add_output('tower_mass_cost_coeff',         units='USD/kg', val=2.9)
        indeps.add_output('controls_machine_rating_cost_coeff', units='USD/kW', val=21.15)
        indeps.add_output('crane_cost',                    units='USD', val=12e3)
        
        indeps.add_output('hub_assemblyCostMultiplier',    val=0.0)
        indeps.add_output('hub_overheadCostMultiplier',    val=0.0)
        indeps.add_output('nacelle_assemblyCostMultiplier',val=0.0)
        indeps.add_output('nacelle_overheadCostMultiplier',val=0.0)
        indeps.add_output('tower_assemblyCostMultiplier',  val=0.0)
        indeps.add_output('tower_overheadCostMultiplier',  val=0.0)
        indeps.add_output('turbine_assemblyCostMultiplier',val=0.0)
        indeps.add_output('turbine_overheadCostMultiplier',val=0.0)
        indeps.add_output('hub_profitMultiplier',          val=0.0)
        indeps.add_output('nacelle_profitMultiplier',      val=0.0)
        indeps.add_output('tower_profitMultiplier',        val=0.0)
        indeps.add_output('turbine_profitMultiplier',      val=0.0)
        indeps.add_output('hub_transportMultiplier',       val=0.0)
        indeps.add_output('nacelle_transportMultiplier',   val=0.0)
        indeps.add_output('tower_transportMultiplier',     val=0.0)
        indeps.add_output('turbine_transportMultiplier',   val=0.0)
        self.add_subsystem('indeps', indeps, promotes=['*'])
        
        self.add_subsystem('blade_c'       , BladeCost2015(),         promotes=['*'])
        self.add_subsystem('hub_c'         , HubCost2015(),           promotes=['*'])
        self.add_subsystem('pitch_c'       , PitchSystemCost2015(),   promotes=['*'])
        self.add_subsystem('spinner_c'     , SpinnerCost2015(),       promotes=['*'])
        self.add_subsystem('hub_adder'     , HubSystemCostAdder2015(),promotes=['*'])
        self.add_subsystem('rotor_adder'   , RotorCostAdder2015(),    promotes=['*'])
        self.add_subsystem('lss_c'         , LowSpeedShaftCost2015(), promotes=['*'])
        self.add_subsystem('bearing_c'     , BearingsCost2015(),      promotes=['*'])
        self.add_subsystem('gearbox_c'     , GearboxCost2015(),       promotes=['*'])
        self.add_subsystem('hss_c'         , HighSpeedSideCost2015(), promotes=['*'])
        self.add_subsystem('generator_c'   , GeneratorCost2015(),     promotes=['*'])
        self.add_subsystem('bedplate_c'    , BedplateCost2015(),      promotes=['*'])
        self.add_subsystem('yaw_c'         , YawSystemCost2015(),     promotes=['*'])
        self.add_subsystem('hvac_c'        , HydraulicCoolingCost2015(), promotes=['*'])
        self.add_subsystem('controls_c'    , ControlsCost2015(),      promotes=['*'])
        self.add_subsystem('vs_c'          , VariableSpeedElecCost2015(), promotes=['*'])
        self.add_subsystem('elec_c'        , ElecConnecCost2015(),    promotes=['*'])
        self.add_subsystem('cover_c'       , NacelleCoverCost2015(),  promotes=['*'])
        self.add_subsystem('other_c'       , OtherMainframeCost2015(),promotes=['*'])
        self.add_subsystem('transformer_c' , TransformerCost2015(),   promotes=['*'])
        self.add_subsystem('nacelle_adder' , NacelleSystemCostAdder2015(), promotes=['*'])
        self.add_subsystem('tower_c'       , TowerCost2015(),         promotes=['*'])
        self.add_subsystem('tower_adder'   , TowerCostAdder2015(),    promotes=['*'])
        self.add_subsystem('turbine_c'     , TurbineCostAdder2015(),  promotes=['*'])
        self.add_subsystem('outputs'       , Outputs2Screen(verbosity=self.verbosity), promotes=['*'])

#-------------------------------------------------------------------------------
def example():

    # simple test of module
    turbine = Turbine_CostsSE_2015(verbosity=True)
    prob = Problem(turbine)
    prob.setup()

    prob['blade_mass']          = 17650.67  # inline with the windpact estimates
    prob['hub_mass']            = 31644.5
    prob['pitch_system_mass']   = 17004.0
    prob['spinner_mass']        = 1810.5
    prob['lss_mass']            = 31257.3
    #bearingsMass'] = 9731.41
    prob['main_bearing_mass']   = 9731.41 / 2
    prob['gearbox_mass']        = 30237.60
    prob['hss_mass']            = 1492.45
    prob['generator_mass']      = 16699.85
    prob['bedplate_mass']       = 93090.6
    prob['yaw_mass']            = 11878.24
    prob['tower_mass']          = 434559.0
    prob['vs_electronics_mass'] = 1000.
    prob['hvac_mass']           = 1000.
    prob['cover_mass']          = 1000.
    prob['platforms_mass']      = 1000.
    prob['transformer_mass']    = 1000.

    # other inputs
    prob['machine_rating']      = 5000.0
    prob['blade_number']        = 3
    prob['crane']               = True
    prob['main_bearing_number'] = 2

    prob.run_model()

    #print('The results for the NREL 5 MW Reference Turbine in an offshore 20 m water depth location are')
    for io in turbine._outputs:
        print(io, str(turbine._outputs[io]))


if __name__ == "__main__":

    example()



#-------------------------------------------------------------------------------
# Name:        RigidMember.Py
# Purpose:     Rigid Member Class
#
# Author:      rdamiani
#
# Created:     12/07/2013
# Copyright:   (c) rdamiani 2013
# Licence:     <your licence>
#-------------------------------------------------------------------------------
from wisdem.commonse.Material import Material

def main():
    pass

if __name__ == '__main__':
    main()

class RigidMember:

    def __init__(self,**kwargs):
        """Rigid member structural virtual properties for FRAME3DD (very high stiffenss\n
        very low density) to simulate TP struts and to connect RNA to top of tower."""
        prms={'name':'~0-weight_rigid','E':10**15,'nu':0.33,'rho':10., 'fy':10**20,'fyc':10**20,'D':1.,'t':0.01,'L':1.,'Kbuck':1.} #SI Units
        prms.update(kwargs) #update in case user put some new params in
        from sys import float_info as CPU
        self.Area=self.Asy=self.Asx=1. #Fix area and shear area to 1 m^2 and then we will increase E and G by 10 OoMs
        self.Jxx=self.Jyy=1. #Again arbitrary set to 1
        self.J0=self.Jyy+self.Jxx
        self.mat=Material(name=prms['name'],E=prms['E'],rho=prms['rho'],nu=prms['nu'],fy=prms['fy'],fyc=prms['fyc']) #rho=CPU.epsilon, E=10**20
        #Auxiliary D and t, L used just as placeholders
        self.D=prms['D']
        self.t=prms['t']
        self.L=prms['L']
        self.Kbuck=prms['Kbuck']



import numpy as np
import numpy.testing as npt
import unittest
import wisdem.floatingse.substructure as subs

from wisdem.commonse import gravity as g
NSECTIONS = 5
NPTS = 100

class TestSubs(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}
        self.resids = {}

        self.inputs['structural_mass'] = 1e4
        self.inputs['structure_center_of_mass'] = 40.0*np.ones(3)
        self.inputs['structural_frequencies'] = 100.0*np.ones(6)
        self.inputs['total_force'] = 26.0*np.ones(3)
        self.inputs['total_moment'] = 5e4*np.ones(3)
        self.inputs['total_displacement'] = 1e4
        
        self.inputs['mooring_mass'] = 20.0
        self.inputs['mooring_neutral_load'] = np.zeros((15,3))
        self.inputs['mooring_neutral_load'][:3,:] = 5.0*g
        self.inputs['mooring_surge_restoring_force'] = 1e2
        self.inputs['mooring_pitch_restoring_force'] = 1e5 * np.ones((10,3))
        self.inputs['mooring_pitch_restoring_force'][3:,:] = 0.0
        self.inputs['mooring_cost'] = 256.0
        self.inputs['mooring_stiffness'] = np.ones((6,6))
        self.inputs['mooring_moments_of_inertia'] = np.array([10.0, 10.0, 2.0, 0.0, 0.0, 0.0])
        self.inputs['fairlead'] = 0.5
        self.inputs['fairlead_location'] = 0.1
        self.inputs['fairlead_radius'] = 5.0
        self.inputs['max_survival_heel'] = 10.0
        self.inputs['operational_heel'] = 10.0

        self.inputs['pontoon_cost'] = 512.0
        

        self.inputs['Hs'] = 10.0
        self.inputs['wave_period'] = 50.0
        
        self.inputs['main_Iwaterplane'] = 150.0
        self.inputs['main_Awaterplane'] = 20.0
        self.inputs['main_mass'] = 2.0*np.ones(NPTS-1)
        self.inputs['main_cost'] = 32.0
        self.inputs['main_freeboard'] = 10.0
        self.inputs['main_center_of_mass'] = -10.0
        self.inputs['main_center_of_buoyancy'] = -8.0
        self.inputs['main_added_mass'] = 2*np.array([10.0, 10.0, 2.0, 30.0, 30.0, 0.0])
        self.inputs['main_moments_of_inertia'] = 1e2 * np.array([10.0, 10.0, 2.0, 0.0, 0.0, 0.0])

        self.inputs['offset_Iwaterplane'] = 50.0
        self.inputs['offset_Awaterplane'] = 9.0
        self.inputs['offset_cost'] = 64.0
        self.inputs['offset_mass'] = np.ones(NPTS-1)
        self.inputs['offset_center_of_mass'] = -5.0
        self.inputs['offset_center_of_buoyancy'] = -4.0
        self.inputs['offset_added_mass'] = np.array([10.0, 10.0, 2.0, 30.0, 30.0, 0.0])
        self.inputs['offset_moments_of_inertia'] = 1e1 * np.array([10.0, 10.0, 2.0, 0.0, 0.0, 0.0])
        self.inputs['offset_freeboard'] = 10.0
        self.inputs['offset_draft'] = 15.0

        self.inputs['tower_z_full'] = np.linspace(0, 90, 3*NSECTIONS+1)
        self.inputs['tower_mass'] = 2e2
        self.inputs['tower_shell_cost'] = 2e5
        self.inputs['tower_d_full'] = 5.0*np.ones(NPTS)
        self.inputs['tower_I_base'] = 1e5*np.array([1.0, 1.0, 1.0, 0.0, 0.0, 0.0])
        self.inputs['rna_mass'] = 6e1
        self.inputs['rna_cg'] = np.array([0.0, 0.0, 5.0])
        self.inputs['rna_I'] = 1e5*np.array([1.0, 1.0, 1.0, 0.0, 0.0, 0.0])
        self.inputs['Rhub'] = 3.0
        
        self.inputs['number_of_offset_columns'] = 3
        self.inputs['water_ballast_radius_vector'] = 40.0 * np.ones(5)
        self.inputs['water_ballast_zpts_vector'] = np.array([-10, -9, -8, -7, -6])
        self.inputs['radius_to_offset_column'] = 20.0
        self.inputs['z_center_of_buoyancy'] = -2.0

        self.inputs['water_density'] = 1e3
        self.inputs['wave_period_range_low'] = 2.0
        self.inputs['wave_period_range_high'] = 20.0

        self.mysemi = subs.Substructure(nFull=NPTS,nFullTow=NPTS)
        self.mysemiG = subs.SubstructureGeometry(nFull=2,nFullTow=2)
        
    def testSetGeometry(self):
        self.inputs['number_of_offset_columns'] = 3
        self.inputs['Rhub'] = 3.0
        self.inputs['tower_d_full'] = 2.0*5.0*np.ones(3)
        self.inputs['main_d_full'] = 2*np.array([10.0, 10.0, 10.0])
        self.inputs['offset_d_full'] = 2*np.array([10.0, 10.0, 10.0])
        self.inputs['offset_z_nodes'] = np.array([-35.0, -15.0, 15.0])
        self.inputs['main_z_nodes'] = np.array([-35.0, -15.0, 15.0])
        self.inputs['radius_to_offset_column'] = 25.0
        self.inputs['fairlead_location'] = 0.1
        self.inputs['fairlead_offset_from_shell'] = 1.0
        self.inputs['offset_freeboard'] = 10.0
        self.inputs['offset_draft'] = 15.0
        self.mysemiG.compute(self.inputs, self.outputs)

        # Semi
        self.assertEqual(self.outputs['fairlead'], 30.0)
        self.assertEqual(self.outputs['fairlead_radius'], 11.0+25.0)
        self.assertEqual(self.outputs['main_offset_spacing'], 25.0 - 10.0 - 10.0)
        self.assertEqual(self.outputs['tower_transition_buffer'], 10-5.0)
        self.assertEqual(self.outputs['nacelle_transition_buffer'], 4.0-5.0)
        self.assertEqual(self.outputs['offset_freeboard_heel_margin'], 10.0 - 25.0*np.sin(np.deg2rad(10.0)))
        self.assertEqual(self.outputs['offset_draft_heel_margin'], 15.0 - 25.0*np.sin(np.deg2rad(10.0)))

        # Spar
        self.inputs['number_of_offset_columns'] = 0
        self.mysemiG.compute(self.inputs, self.outputs)
        self.assertEqual(self.outputs['fairlead'], 30.0)
        self.assertEqual(self.outputs['fairlead_radius'], 11.0)
        self.assertEqual(self.outputs['main_offset_spacing'], 25.0 - 10.0 - 10.0)
        self.assertEqual(self.outputs['tower_transition_buffer'], 10-5.0)
        self.assertEqual(self.outputs['nacelle_transition_buffer'], 4.0-5.0)

        
    def testBalance(self):
        self.mysemi.balance(self.inputs, self.outputs)
        m_water = 1e3*1e4 - 1e4 - 15
        z_data = self.inputs['water_ballast_zpts_vector']
        h_data = z_data - z_data[0]
        h_expect = np.interp(m_water, 1e3*h_data*np.pi*self.inputs['water_ballast_radius_vector']**2, h_data)
        cg_expect_z = (1e4*40.0 + m_water*(-10 + 0.5*h_expect)) / (1e4+m_water)
        cg_expect_xy = 1e4*40.0/ (1e4+m_water)

        self.assertEqual(self.outputs['variable_ballast_mass'], m_water)
        self.assertEqual(self.outputs['variable_ballast_height_ratio'], h_expect/4.0)
        npt.assert_almost_equal(self.outputs['center_of_mass'], np.array([cg_expect_xy, cg_expect_xy, cg_expect_z]))
        
        self.inputs['number_of_offset_columns'] = 0
        self.mysemi.balance(self.inputs, self.outputs)

        self.assertEqual(self.outputs['variable_ballast_mass'], m_water)
        self.assertEqual(self.outputs['variable_ballast_height_ratio'], h_expect/4.0)
        npt.assert_almost_equal(self.outputs['center_of_mass'], np.array([cg_expect_xy, cg_expect_xy, cg_expect_z]))
        

    def testStability(self):
        self.inputs['mooring_pitch_restoring_force'] = 0.0 * np.ones((10,3))
        self.outputs['center_of_mass'] = np.array([0.0, 0.0, -1.0])
        self.mysemi.compute_stability(self.inputs, self.outputs)

        I_expect = 150.0 + (50.0 + 9.0*(20.0*np.cos(np.deg2rad(np.array([0.0, 120., 240.0]))) )**2).sum()
        static_expect = -1.0 + 2.0
        meta_expect = I_expect/1e4 - static_expect
        wind_fact = np.cos(np.deg2rad(10.0))**2.0
        self.assertEqual(self.outputs['buoyancy_to_gravity'], static_expect)
        self.assertEqual(self.outputs['metacentric_height'], meta_expect)
        self.assertEqual(self.outputs['offset_force_ratio'], 26.0/1e2)
        self.assertAlmostEqual(self.outputs['heel_moment_ratio'], (wind_fact*5e4)/(1e4*g*1e3*np.sin(np.deg2rad(10))*np.abs(meta_expect)))

        self.inputs['number_of_offset_columns'] = 0
        self.mysemi.compute_stability(self.inputs, self.outputs)

        I_expect = 150.0
        meta_expect = I_expect/1e4 - static_expect
        self.assertEqual(self.outputs['buoyancy_to_gravity'], static_expect)
        self.assertEqual(self.outputs['metacentric_height'], meta_expect)
        self.assertEqual(self.outputs['offset_force_ratio'], 26.0/1e2)
        self.assertAlmostEqual(self.outputs['heel_moment_ratio'], (wind_fact*5e4)/(1e4*g*1e3*np.sin(np.deg2rad(10))*np.abs(meta_expect)))

        self.inputs['fairlead'] = 1.0
        self.inputs['mooring_pitch_restoring_force'][:3,-1] = 1.0
        self.assertAlmostEqual(self.outputs['heel_moment_ratio'], (wind_fact*5e4)/(1*5 + 1e4*g*1e3*np.sin(np.deg2rad(10))*np.abs(meta_expect)))


    def testPeriods(self):
        # Spar first
        self.inputs['structure_center_of_mass'] = np.array([0.0, 0.0, -40.0])
        self.inputs['number_of_offset_columns'] = 0
        self.mysemi.balance(self.inputs, self.outputs)
        self.inputs['main_center_of_mass'] = self.outputs['center_of_mass'][-1]
        self.inputs['main_center_of_buoyancy'] = self.outputs['center_of_mass'][-1]+2.0
        self.inputs['tower_mass'] = 0.0
        self.inputs['rna_mass'] = 0.0
        self.inputs['rna_I'] = 1e2*np.array([1.0, 1.0, 1.0, 0.0, 0.0, 0.0])
        self.inputs['tower_I_base'] = 1e2*np.array([1.0, 1.0, 1.0, 0.0, 0.0, 0.0])
        self.mysemi.compute_stability(self.inputs, self.outputs)
        self.mysemi.compute_rigid_body_periods(self.inputs, self.outputs)

        m_struct = self.inputs['structural_mass']
        m_water  = self.outputs['variable_ballast_mass']
        z_cg     = self.inputs['main_center_of_mass']
        z_water  = self.outputs['variable_ballast_center_of_mass']
        I_water  = self.outputs['variable_ballast_moments_of_inertia']
        M_expect = np.zeros(6)
        M_expect[:3] = m_struct + m_water
        M_expect[3:] = self.inputs['main_moments_of_inertia'][:3]
        M_expect[3:5] += I_water[:2] + m_water*(z_water-z_cg)**2
        M_expect[-1] += I_water[2]
        M_expect[3:] += 2e2
        npt.assert_equal(self.outputs['mass_matrix'], M_expect)

        A_expect = np.zeros(6)
        A_expect[:3] = self.inputs['main_added_mass'][:3]
        A_expect[3:5] = self.inputs['main_added_mass'][3:5] + A_expect[0]*(2.0)**2
        npt.assert_equal(self.outputs['added_mass_matrix'], A_expect)

        rho_w = self.inputs['water_density']
        K_expect = np.zeros(6)
        K_expect[2] = rho_w * g * 20.0 # waterplane area for main
        K_expect[3:5] = rho_w * g * self.outputs['metacentric_height'] * 1e4 # Total displacement
        npt.assert_almost_equal(self.outputs['hydrostatic_stiffness'], K_expect)

        T_expect = 2 * np.pi * np.sqrt( (M_expect + A_expect) / (1e-6 + K_expect + np.diag(self.inputs['mooring_stiffness'])) )
        npt.assert_almost_equal(self.outputs['rigid_body_periods'], T_expect)

        
    def testMargins(self):
        
        self.mysemi.balance(self.inputs, self.outputs)
        self.mysemi.compute_stability(self.inputs, self.outputs)
        self.mysemi.compute_rigid_body_periods(self.inputs, self.outputs)
        self.mysemi.check_frequency_margins(self.inputs, self.outputs)

        myones = np.ones(6)
        T_sys    = self.outputs['rigid_body_periods']
        T_wave_low  = self.inputs['wave_period_range_low']*myones
        T_wave_high = self.inputs['wave_period_range_high']*myones
        f_struct    = self.inputs['structural_frequencies']
        T_struct    = 1.0 / f_struct

        T_wave_high[-1] = 1e-16
        T_wave_low[-1] = 1e30

        ind = T_sys>T_wave_low
        npt.assert_equal(self.outputs['period_margin_high'][ind], T_sys[ind]/T_wave_high[ind] )

        ind = T_sys<T_wave_high
        npt.assert_equal(self.outputs['period_margin_low'][ind], T_sys[ind]/T_wave_low[ind] )

        ind = T_struct>T_wave_low
        npt.assert_equal(self.outputs['modal_margin_high'][ind], T_struct[ind]/T_wave_high[ind] )

        ind = T_struct<T_wave_high
        npt.assert_equal(self.outputs['modal_margin_low'][ind], T_struct[ind]/T_wave_low[ind] )
        
    def testCost(self):
        self.mysemi.compute_costs(self.inputs, self.outputs)
        c_expect = 256.0 + 512.0 + 32.0 + 3*64.0 + 2e5
        self.assertEqual(self.outputs['total_cost'], c_expect)

        
def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestSubs))
    return suite

if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())
    
    



from __future__ import print_function

import numpy as np
from scipy.optimize import curve_fit
from scipy.interpolate import PchipInterpolator
import os, copy, warnings, shutil
from openmdao.api import IndepVarComp, ExplicitComponent, Group, Problem
from wisdem.commonse.mpi_tools import MPI

from wisdem.aeroelasticse.FAST_reader import InputReader_Common, InputReader_OpenFAST, InputReader_FAST7
from wisdem.aeroelasticse.FAST_writer import InputWriter_Common, InputWriter_OpenFAST, InputWriter_FAST7
from wisdem.aeroelasticse.FAST_wrapper import FastWrapper
from wisdem.aeroelasticse.runFAST_pywrapper import runFAST_pywrapper, runFAST_pywrapper_batch
from wisdem.aeroelasticse.CaseLibrary import RotorSE_rated, RotorSE_DLC_1_4_Rated, RotorSE_DLC_7_1_Steady, RotorSE_DLC_1_1_Turb, power_curve
from wisdem.aeroelasticse.FAST_post import return_timeseries


if MPI:
    #from openmdao.api import PetscImpl as impl
    from mpi4py import MPI
    from petsc4py import PETSc
# else:
    # from openmdao.api import BasicImpl as impl



def eval_unsteady(alpha, cl, cd, cm):
    # calculate unsteady coefficients from polars for OpenFAST's Aerodyn

    unsteady = {}

    alpha_rad = np.radians(alpha)
    cn = cl*np.cos(alpha_rad) + cd*np.sin(alpha_rad)

    # alpha0, Cd0, Cm0
    aoa_l = [-30.]
    aoa_h = [30.]
    idx_low  = np.argmin(abs(alpha-aoa_l))
    idx_high = np.argmin(abs(alpha-aoa_h))

    if max(np.abs(np.gradient(cl)))>0.:
        unsteady['alpha0'] = np.interp(0., cl[idx_low:idx_high], alpha[idx_low:idx_high])
        unsteady['Cd0'] = np.interp(0., cl[idx_low:idx_high], cd[idx_low:idx_high])
        unsteady['Cm0'] = np.interp(0., cl[idx_low:idx_high], cm[idx_low:idx_high])
    else:
        unsteady['alpha0'] = 0.
        unsteady['Cd0'] = cd[np.argmin(abs(alpha-0.))]
        unsteady['Cm0'] = 0.


    unsteady['eta_e']= 1
    unsteady['T_f0'] = "Default"
    unsteady['T_V0'] = "Default"
    unsteady['T_p']  = "Default"
    unsteady['T_VL'] = "Default"
    unsteady['b1']   = "Default"
    unsteady['b2']   = "Default"
    unsteady['b5']   = "Default"
    unsteady['A1']   = "Default"
    unsteady['A2']   = "Default"
    unsteady['A5']   = "Default"
    unsteady['S1']   = 0
    unsteady['S2']   = 0
    unsteady['S3']   = 0
    unsteady['S4']   = 0

    def find_breakpoint(x, y, idx_low, idx_high, multi=1.):
        lin_fit = np.interp(x[idx_low:idx_high], [x[idx_low],x[idx_high]], [y[idx_low],y[idx_high]])
        idx_break = 0
        lin_diff = 0
        for i, (fit, yi) in enumerate(zip(lin_fit, y[idx_low:idx_high])):
            if multi==0:
                diff_i = np.abs(yi-fit)
            else:
                diff_i = multi*(yi-fit)
            if diff_i>lin_diff:
                lin_diff = diff_i
                idx_break = i
        idx_break += idx_low
        return idx_break

    # Cn1
    idx_alpha0  = np.argmin(abs(alpha-unsteady['alpha0']))
    
    if max(np.abs(np.gradient(cm)))>1.e-10:
        aoa_h = alpha[idx_alpha0]+35.
        idx_high = np.argmin(abs(alpha-aoa_h))

        cm_temp = cm[idx_low:idx_high]
        idx_cm_min = [i for i,local_min in enumerate(np.r_[True, cm_temp[1:] < cm_temp[:-1]] & np.r_[cm_temp[:-1] < cm_temp[1:], True]) if local_min] + idx_low
        idx_high = idx_cm_min[-1]
        
        
        idx_Cn1 = find_breakpoint(alpha, cm, idx_alpha0, idx_high)
        unsteady['Cn1'] = cn[idx_Cn1]
    else:
        idx_Cn1 = np.argmin(abs(alpha-0.))
        unsteady['Cn1'] = 0.
    

    
    # Cn2
    if max(np.abs(np.gradient(cm)))>1.e-10:
        aoa_l = np.mean([alpha[idx_alpha0], alpha[idx_Cn1]])-30.
        idx_low  = np.argmin(abs(alpha-aoa_l))

        cm_temp = cm[idx_low:idx_high]
        idx_cm_min = [i for i,local_min in enumerate(np.r_[True, cm_temp[1:] < cm_temp[:-1]] & np.r_[cm_temp[:-1] < cm_temp[1:], True]) if local_min] + idx_low
        idx_high = idx_cm_min[-1]
        
        idx_Cn2 = find_breakpoint(alpha, cm, idx_low, idx_alpha0, multi=0.)
        unsteady['Cn2'] = cn[idx_Cn2]
    else:
        idx_Cn2 = np.argmin(abs(alpha-0.))
        unsteady['Cn2'] = 0.

    # C_nalpha
    if max(np.abs(np.gradient(cm)))>1.e-10:
        # unsteady['C_nalpha'] = np.gradient(cn, alpha_rad)[idx_alpha0]
        unsteady['C_nalpha'] = max(np.gradient(cn[idx_alpha0:idx_Cn1], alpha_rad[idx_alpha0:idx_Cn1]))

    else:
        unsteady['C_nalpha'] = 0.

    # alpha1, alpha2
    # finding the break point in drag as a proxy for Trailing Edge separation, f=0.7
    # 3d stall corrections cause erroneous f calculations 
    if max(np.abs(np.gradient(cm)))>1.0e-10:
        aoa_l = [0.]
        idx_low  = np.argmin(abs(alpha-aoa_l))
        idx_alpha1 = find_breakpoint(alpha, cd, idx_low, idx_Cn1, multi=-1.)
        unsteady['alpha1'] = alpha[idx_alpha1]
    else:
        idx_alpha1 = np.argmin(abs(alpha-0.))
        unsteady['alpha1'] = 0.
    unsteady['alpha2'] = -1.*unsteady['alpha1']


    unsteady['St_sh']   = "Default"
    unsteady['k0']      = 0
    unsteady['k1']      = 0
    unsteady['k2']      = 0
    unsteady['k3']      = 0
    unsteady['k1_hat']  = 0
    unsteady['x_cp_bar']   = "Default"
    unsteady['UACutout']   = "Default"
    unsteady['filtCutOff'] = "Default"

    unsteady['Alpha']    = alpha
    unsteady['Cl']    = cl
    unsteady['Cd']    = cd
    unsteady['Cm']    = cm

    return unsteady

    # import matplotlib.pyplot as plt
    # fig, ax = plt.subplots(nrows=3, ncols=1, figsize=(6., 8.), sharex=True)
    # ax[0].plot(alpha, cn)
    # ax[0].plot(alpha, cl, '--')
    # ax[0].plot(unsteady['alpha0'], 0.,'o')
    # ax[0].annotate('alpha0', (unsteady['alpha0'], 0.))
    # ax[0].plot(alpha[idx_alpha0], cn[idx_alpha0],'o')
    # ax[0].annotate('C_nalpha', (alpha[idx_alpha0], cn[idx_alpha0]))
    # ax[0].plot(alpha[idx_Cn1], cn[idx_Cn1],'o')
    # ax[0].annotate('Cn1', (alpha[idx_Cn1], cn[idx_Cn1]))
    # ax[0].plot(alpha[idx_Cn2], cn[idx_Cn2],'o')
    # ax[0].annotate('Cn2', (alpha[idx_Cn2], cn[idx_Cn2]))
    # ax[0].set_ylabel('C_L')
    # ax[0].grid(True, linestyle=':')

    # ax[1].plot(alpha, cd)
    # ax[1].set_ylabel('C_D')
    # ax[1].grid(True, linestyle=':')

    # ax[2].plot(alpha, cm)
    # ax[2].plot(alpha[idx_Cn1], cm[idx_Cn1], 'o')
    # ax[2].annotate('Cn1', (alpha[idx_Cn1], cm[idx_Cn1]))
    # ax[2].plot(alpha[idx_Cn2], cm[idx_Cn2], 'o')
    # ax[2].annotate('Cn2', (alpha[idx_Cn2], cm[idx_Cn2]))

    # ax[2].set_ylabel('C_M')
    # ax[2].set_xlabel('Angle of Attack, deg')
    # ax[2].grid(True, linestyle=':')

    # plt.show()










class FASTLoadCases(ExplicitComponent):
    def initialize(self):
        self.options.declare('RefBlade')
        self.options.declare('npts_coarse_power_curve', default=20)
        self.options.declare('npts_spline_power_curve', default=200)
        self.options.declare('FASTpref',default={})

    def setup(self):
        RefBlade                = self.options['RefBlade']
        npts_coarse_power_curve = self.options['npts_coarse_power_curve']
        npts_spline_power_curve = self.options['npts_spline_power_curve']
        FASTpref                = self.options['FASTpref']
        NPTS                    = len(RefBlade['pf']['s'])
        n_aoa_grid              = len(RefBlade['airfoils_aoa'])
        n_Re_grid               = len(RefBlade['airfoils_Re'])
        
        self.add_discrete_input('fst_vt_in', val={})

        # ElastoDyn Inputs
        # Assuming the blade modal damping to be unchanged. Cannot directly solve from the Rayleigh Damping without making assumptions. J.Jonkman recommends 2-3% https://wind.nrel.gov/forum/wind/viewtopic.php?t=522
        self.add_input('r',                     val=np.zeros(NPTS), units='m', desc='radial positions. r[0] should be the hub location \
            while r[-1] should be the blade tip. Any number \
            of locations can be specified between these in ascending order.')
        self.add_input('le_location',           val=np.zeros(NPTS), desc='Leading-edge positions from a reference blade axis (usually blade pitch axis). Locations are normalized by the local chord length. Positive in -x direction for airfoil-aligned coordinate system')
        self.add_input('beam:Tw_iner',          val=np.zeros(NPTS), units='m', desc='y-distance to elastic center from point about which above structural properties are computed')
        self.add_input('beam:rhoA',             val=np.zeros(NPTS), units='kg/m', desc='mass per unit length')
        self.add_input('beam:EIyy',             val=np.zeros(NPTS), units='N*m**2', desc='flatwise stiffness (bending about y-direction of airfoil aligned coordinate system)')
        self.add_input('beam:EIxx',             val=np.zeros(NPTS), units='N*m**2', desc='edgewise stiffness (bending about :ref:`x-direction of airfoil aligned coordinate system <blade_airfoil_coord>`)')
        self.add_input('modes_coef_curvefem',   val=np.zeros((3, 5)), desc='mode shapes as 6th order polynomials, in the format accepted by ElastoDyn, [[c_x2, c_],..]')

        # AeroDyn Inputs
        self.add_input('z_az',              val=np.zeros(NPTS), units='m', desc='dimensional aerodynamic grid')
        self.add_input('chord',             val=np.zeros(NPTS), units='m', desc='chord at airfoil locations')
        self.add_input('theta',             val=np.zeros(NPTS), units='deg', desc='twist at airfoil locations')
        self.add_input('precurve',          val=np.zeros(NPTS), units='m', desc='precurve at airfoil locations')
        self.add_input('presweep',          val=np.zeros(NPTS), units='m', desc='presweep at structural locations')
        self.add_input('rthick',            val=np.zeros(NPTS), desc='relative thickness of airfoil distribution')
        self.add_input('Rhub',              val=0.0, units='m', desc='dimensional radius of hub')
        self.add_input('Rtip',              val=0.0, units='m', desc='dimensional radius of tip')
        self.add_input('airfoils_cl',       val=np.zeros((n_aoa_grid, NPTS, n_Re_grid)), desc='lift coefficients, spanwise')
        self.add_input('airfoils_cd',       val=np.zeros((n_aoa_grid, NPTS, n_Re_grid)), desc='drag coefficients, spanwise')
        self.add_input('airfoils_cm',       val=np.zeros((n_aoa_grid, NPTS, n_Re_grid)), desc='moment coefficients, spanwise')
        self.add_input('airfoils_aoa',      val=np.zeros((n_aoa_grid)), units='deg', desc='angle of attack grid for polars')
        self.add_input('airfoils_Re',       val=np.zeros((n_Re_grid)), desc='Reynolds numbers of polars')
        
        # Airfoil coordinates
        self.add_input('airfoils_coord_x',  val=np.zeros((200, NPTS)), desc='x airfoil coordinate, spanwise')
        self.add_input('airfoils_coord_y',  val=np.zeros((200, NPTS)), desc='y airfoil coordinate, spanwise')
        
        # Turbine level inputs
        self.add_input('hub_height',                val=0.0, units='m', desc='hub height')
        self.add_discrete_input('turbulence_class', val='A', desc='IEC turbulence class')
        self.add_discrete_input('turbine_class',    val='I', desc='IEC turbulence class')
        self.add_input('control_ratedPower',        val=0.,  units='W',    desc='machine power rating')
        self.add_input('control_maxOmega',          val=0.0, units='rpm',  desc='maximum allowed rotor rotation speed')
        self.add_input('control_maxTS',             val=0.0, units='m/s',  desc='maximum allowed blade tip speed')

        # Initial conditions
        self.add_input('U_init',        val=np.zeros(npts_coarse_power_curve), units='m/s', desc='wind speeds')
        self.add_input('Omega_init',    val=np.zeros(npts_coarse_power_curve), units='rpm', desc='rotation speeds to run')
        self.add_input('pitch_init',    val=np.zeros(npts_coarse_power_curve), units='deg', desc='pitch angles to run')
        self.add_input('V_out',         val=np.zeros(npts_spline_power_curve), units='m/s', desc='wind speeds to output powercurve')
        self.add_input('V',             val=np.zeros(npts_coarse_power_curve), units='m/s',  desc='wind vector')


        # Environmental conditions 
        self.add_input('Vrated',    val=11.0, units='m/s', desc='rated wind speed')
        self.add_input('V_R25',     val=0.0, units='m/s', desc='region 2.5 transition wind speed')
        self.add_input('Vgust',     val=11.0, units='m/s', desc='gust wind speed')
        self.add_input('Vextreme',  val=11.0, units='m/s', desc='IEC extreme wind speed at hub height')
        self.add_input('V_mean_iec', val=11.0, units='m/s', desc='IEC mean wind for turbulence class')
        self.add_input('rho',       val=0.0,        units='kg/m**3',    desc='density of air')
        self.add_input('mu',        val=0.0,        units='kg/(m*s)',   desc='dynamic viscosity of air')
        self.add_input('shearExp',  val=0.0,                            desc='shear exponent')

        # FAST run preferences
        self.FASTpref            = FASTpref 
        self.Analysis_Level      = FASTpref['Analysis_Level']
        self.FAST_ver            = FASTpref['FAST_ver']
        self.FAST_exe            = os.path.abspath(FASTpref['FAST_exe'])
        self.FAST_directory      = os.path.abspath(FASTpref['FAST_directory'])
        self.Turbsim_exe         = os.path.abspath(FASTpref['Turbsim_exe'])
        self.debug_level         = FASTpref['debug_level']
        self.FAST_InputFile      = FASTpref['FAST_InputFile']
        if MPI:
            self.FAST_runDirectory = os.path.join(FASTpref['FAST_runDirectory'],'rank_%000d'%int(impl.world_comm().rank))
            self.FAST_namingOut  = FASTpref['FAST_namingOut']+'_%000d'%int(impl.world_comm().rank)
            # try:
            #     if not os.path.exists(directory):
            #         os.makedirs(self.FAST_runDirectory)
            # except:
            #     pass
        else:
            self.FAST_runDirectory = FASTpref['FAST_runDirectory']
            self.FAST_namingOut  = FASTpref['FAST_namingOut']
        self.dev_branch          = FASTpref['dev_branch']
        self.cores               = FASTpref['cores']
        self.case                = {}
        self.channels            = {}

        # DLC Flags
        self.DLC_powercurve      = FASTpref['DLC_powercurve']
        self.DLC_gust            = FASTpref['DLC_gust']
        self.DLC_extrm           = FASTpref['DLC_extrm']
        self.DLC_turbulent       = FASTpref['DLC_turbulent']

        self.clean_FAST_directory = False
        if 'clean_FAST_directory' in FASTpref.keys():
            self.clean_FAST_directory = FASTpref['clean_FAST_directory']

        self.mpi_run             = False
        if 'mpi_run' in FASTpref.keys():
            self.mpi_run         = FASTpref['mpi_run']
            if self.mpi_run:
                self.mpi_comm_map_down   = FASTpref['mpi_comm_map_down']
        
        self.add_output('dx_defl', val=0., desc='deflection of blade section in airfoil x-direction under max deflection loading')
        self.add_output('dy_defl', val=0., desc='deflection of blade section in airfoil y-direction under max deflection loading')
        self.add_output('dz_defl', val=0., desc='deflection of blade section in airfoil z-direction under max deflection loading')
    
        self.add_output('root_bending_moment', val=0.0, units='N*m', desc='total magnitude of bending moment at root of blade 1')
        self.add_output('Mxyz',         val=np.array([0.0, 0.0, 0.0]), units='N*m', desc='individual moments [x,y,z] at the blade root in blade c.s.')
        
        self.add_output('loads_r',      val=np.zeros(NPTS), units='m', desc='radial positions along blade going toward tip')
        self.add_output('loads_Px',     val=np.zeros(NPTS), units='N/m', desc='distributed loads in blade-aligned x-direction')
        self.add_output('loads_Py',     val=np.zeros(NPTS), units='N/m', desc='distributed loads in blade-aligned y-direction')
        self.add_output('loads_Pz',     val=np.zeros(NPTS), units='N/m', desc='distributed loads in blade-aligned z-direction')
        self.add_output('loads_Omega',  val=0.0, units='rpm', desc='rotor rotation speed')
        self.add_output('loads_pitch',  val=0.0, units='deg', desc='pitch angle')
        self.add_output('loads_azimuth', val=0.0, units='deg', desc='azimuthal angle')
        self.add_discrete_output('model_updated', val=False, desc='boolean, Analysis Level 0: fast model written, but not run')
        self.add_discrete_output('FASTpref_updated', val={}, desc='updated fast preference dictionary')

        self.add_output('P_out',        val=np.zeros(npts_spline_power_curve), units='W', desc='electrical power from rotor')
        self.add_output('P',            val=np.zeros(npts_coarse_power_curve), units='W',    desc='rotor electrical power')
        self.add_output('Cp',           val=np.zeros(npts_coarse_power_curve),               desc='rotor electrical power coefficient')
        self.add_output('rated_V',      val=0.0, units='m/s', desc='rated wind speed')
        self.add_output('rated_Omega',  val=0.0, units='rpm', desc='rotor rotation speed at rated')
        self.add_output('rated_pitch',  val=0.0, units='deg', desc='pitch setting at rated')
        self.add_output('rated_T',      val=0.0, units='N', desc='rotor aerodynamic thrust at rated')
        self.add_output('rated_Q',      val=0.0, units='N*m', desc='rotor aerodynamic torque at rated')

        self.add_discrete_output('fst_vt_out', val={})

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        #print(impl.world_comm().rank, 'Rotor_fast','start')

        fst_vt, R_out = self.update_FAST_model(inputs, discrete_inputs)
        
        # if MPI:
            # rank = int(PETSc.COMM_WORLD.getRank())
            # self.FAST_namingOut = self.FAST_namingOut + '_%00d'%rank

        if self.Analysis_Level == 2:
            # Run FAST with ElastoDyn
            list_cases, list_casenames, required_channels, case_keys = self.DLC_creation(inputs, discrete_inputs, fst_vt)
            FAST_Output = self.run_FAST(fst_vt, list_cases, list_casenames, required_channels)
            self.post_process(FAST_Output, case_keys, R_out, inputs, discrete_inputs, outputs)

        elif self.Analysis_Level == 1:
            # Write FAST files, do not run
            self.write_FAST(fst_vt, discrete_outputs)

        discrete_outputs['fst_vt_out'] = fst_vt

        # delete run directory. not recommended for most cases, use for large parallelization problems where disk storage will otherwise fill up
        if self.clean_FAST_directory:
            try:
                shutil.rmtree(self.FAST_runDirectory)
            except:
                print('Failed to delete directory: %s'%self.FAST_runDirectory)

        #print(impl.world_comm().rank, 'Rotor_fast','end')


    def update_FAST_model(self, inputs, discrete_inputs):

        # Create instance of FAST reference model 

        fst_vt = copy.deepcopy(discrete_inputs['fst_vt_in'])

        fst_vt['Fst']['OutFileFmt'] = 2

        # Update ElastoDyn
        fst_vt['ElastoDyn']['TipRad'] = inputs['Rtip'][0]
        fst_vt['ElastoDyn']['HubRad'] = inputs['Rhub'][0]
        tower2hub = fst_vt['InflowWind']['RefHt'] - fst_vt['ElastoDyn']['TowerHt']
        fst_vt['ElastoDyn']['TowerHt'] = inputs['hub_height'][0]

        # Update Inflowwind
        fst_vt['InflowWind']['RefHt'] = inputs['hub_height'][0]
        fst_vt['InflowWind']['PLexp'] = inputs['shearExp'][0]

        # Update ElastoDyn Blade Input File
        fst_vt['ElastoDynBlade']['NBlInpSt']   = len(inputs['r'])
        fst_vt['ElastoDynBlade']['BlFract']    = (inputs['r']-inputs['Rhub'])/(inputs['Rtip']-inputs['Rhub'])
        fst_vt['ElastoDynBlade']['BlFract'][0] = 0.
        fst_vt['ElastoDynBlade']['BlFract'][-1]= 1.
        fst_vt['ElastoDynBlade']['PitchAxis']  = inputs['le_location']
        # fst_vt['ElastoDynBlade']['StrcTwst']   = inputs['beam:Tw_iner']
        fst_vt['ElastoDynBlade']['StrcTwst']   = inputs['theta'] # to do: structural twist is not nessessarily (nor likely to be) the same as aero twist
        fst_vt['ElastoDynBlade']['BMassDen']   = inputs['beam:rhoA']
        fst_vt['ElastoDynBlade']['FlpStff']    = inputs['beam:EIyy']
        fst_vt['ElastoDynBlade']['EdgStff']    = inputs['beam:EIxx']
        for i in range(5):
            fst_vt['ElastoDynBlade']['BldFl1Sh'][i] = inputs['modes_coef_curvefem'][0,i]
            fst_vt['ElastoDynBlade']['BldFl2Sh'][i] = inputs['modes_coef_curvefem'][1,i]
            fst_vt['ElastoDynBlade']['BldEdgSh'][i] = inputs['modes_coef_curvefem'][2,i]
        
        # Update AeroDyn15
        fst_vt['AeroDyn15']['AirDens'] = inputs['rho'][0]
        fst_vt['AeroDyn15']['KinVisc'] = inputs['mu'][0] / inputs['rho'][0]
        
        # Update AeroDyn15 Blade Input File
        r = (inputs['r']-inputs['Rhub'])
        r[0]  = 0.
        r[-1] = inputs['Rtip']-inputs['Rhub']
        fst_vt['AeroDynBlade']['NumBlNds'] = len(r)
        fst_vt['AeroDynBlade']['BlSpn']    = r
        fst_vt['AeroDynBlade']['BlCrvAC']  = inputs['precurve']
        fst_vt['AeroDynBlade']['BlSwpAC']  = inputs['presweep']
        fst_vt['AeroDynBlade']['BlCrvAng'] = np.degrees(np.arcsin(np.gradient(inputs['precurve'])/np.gradient(r)))
        fst_vt['AeroDynBlade']['BlTwist']  = inputs['theta']
        fst_vt['AeroDynBlade']['BlChord']  = inputs['chord']
        fst_vt['AeroDynBlade']['BlAFID']   = np.asarray(range(1,len(r)+1))

        # Update AeroDyn15 Airfoile Input Files
        # airfoils = inputs['airfoils']
        fst_vt['AeroDyn15']['NumAFfiles'] = len(r)
        # fst_vt['AeroDyn15']['af_data'] = [{}]*len(airfoils)
        fst_vt['AeroDyn15']['af_data'] = []
        for i in range(len(r)):
            # af = airfoils[i]
            unsteady = eval_unsteady(inputs['airfoils_aoa'], inputs['airfoils_cl'][:,i,0], inputs['airfoils_cd'][:,i,0], inputs['airfoils_cm'][:,i,0])
            fst_vt['AeroDyn15']['af_data'].append({})
            fst_vt['AeroDyn15']['af_data'][i]['InterpOrd'] = "DEFAULT"
            fst_vt['AeroDyn15']['af_data'][i]['NonDimArea']= 1
            fst_vt['AeroDyn15']['af_data'][i]['NumCoords'] = 0          # TODO: link the airfoil profiles to this component and write the coordinate files (no need as of yet)
            fst_vt['AeroDyn15']['af_data'][i]['NumTabs']   = 1
            fst_vt['AeroDyn15']['af_data'][i]['Re']        = 0.75       # TODO: functionality for multiple Re tables
            fst_vt['AeroDyn15']['af_data'][i]['Ctrl']      = 0
            fst_vt['AeroDyn15']['af_data'][i]['InclUAdata']= "True"
            fst_vt['AeroDyn15']['af_data'][i]['alpha0']    = unsteady['alpha0']
            fst_vt['AeroDyn15']['af_data'][i]['alpha1']    = unsteady['alpha1']
            fst_vt['AeroDyn15']['af_data'][i]['alpha2']    = unsteady['alpha2']
            fst_vt['AeroDyn15']['af_data'][i]['eta_e']     = unsteady['eta_e']
            fst_vt['AeroDyn15']['af_data'][i]['C_nalpha']  = unsteady['C_nalpha']
            fst_vt['AeroDyn15']['af_data'][i]['T_f0']      = unsteady['T_f0']
            fst_vt['AeroDyn15']['af_data'][i]['T_V0']      = unsteady['T_V0']
            fst_vt['AeroDyn15']['af_data'][i]['T_p']       = unsteady['T_p']
            fst_vt['AeroDyn15']['af_data'][i]['T_VL']      = unsteady['T_VL']
            fst_vt['AeroDyn15']['af_data'][i]['b1']        = unsteady['b1']
            fst_vt['AeroDyn15']['af_data'][i]['b2']        = unsteady['b2']
            fst_vt['AeroDyn15']['af_data'][i]['b5']        = unsteady['b5']
            fst_vt['AeroDyn15']['af_data'][i]['A1']        = unsteady['A1']
            fst_vt['AeroDyn15']['af_data'][i]['A2']        = unsteady['A2']
            fst_vt['AeroDyn15']['af_data'][i]['A5']        = unsteady['A5']
            fst_vt['AeroDyn15']['af_data'][i]['S1']        = unsteady['S1']
            fst_vt['AeroDyn15']['af_data'][i]['S2']        = unsteady['S2']
            fst_vt['AeroDyn15']['af_data'][i]['S3']        = unsteady['S3']
            fst_vt['AeroDyn15']['af_data'][i]['S4']        = unsteady['S4']
            fst_vt['AeroDyn15']['af_data'][i]['Cn1']       = unsteady['Cn1']
            fst_vt['AeroDyn15']['af_data'][i]['Cn2']       = unsteady['Cn2']
            fst_vt['AeroDyn15']['af_data'][i]['St_sh']     = unsteady['St_sh']
            fst_vt['AeroDyn15']['af_data'][i]['Cd0']       = unsteady['Cd0']
            fst_vt['AeroDyn15']['af_data'][i]['Cm0']       = unsteady['Cm0']
            fst_vt['AeroDyn15']['af_data'][i]['k0']        = unsteady['k0']
            fst_vt['AeroDyn15']['af_data'][i]['k1']        = unsteady['k1']
            fst_vt['AeroDyn15']['af_data'][i]['k2']        = unsteady['k2']
            fst_vt['AeroDyn15']['af_data'][i]['k3']        = unsteady['k3']
            fst_vt['AeroDyn15']['af_data'][i]['k1_hat']    = unsteady['k1_hat']
            fst_vt['AeroDyn15']['af_data'][i]['x_cp_bar']  = unsteady['x_cp_bar']
            fst_vt['AeroDyn15']['af_data'][i]['UACutout']  = unsteady['UACutout']
            fst_vt['AeroDyn15']['af_data'][i]['filtCutOff']= unsteady['filtCutOff']
            fst_vt['AeroDyn15']['af_data'][i]['NumAlf']    = len(unsteady['Alpha'])
            fst_vt['AeroDyn15']['af_data'][i]['Alpha']     = np.array(unsteady['Alpha'])
            fst_vt['AeroDyn15']['af_data'][i]['Cl']        = np.array(unsteady['Cl'])
            fst_vt['AeroDyn15']['af_data'][i]['Cd']        = np.array(unsteady['Cd'])
            fst_vt['AeroDyn15']['af_data'][i]['Cm']        = np.array(unsteady['Cm'])
            fst_vt['AeroDyn15']['af_data'][i]['Cpmin']     = np.zeros_like(unsteady['Cm'])
        
        
        fst_vt['AeroDyn15']['af_coord'] = []
        fst_vt['AeroDyn15']['rthick']   = np.zeros(len(r))
        for i in range(len(r)):
            fst_vt['AeroDyn15']['af_coord'].append({})
            fst_vt['AeroDyn15']['af_coord'][i]['x']  = inputs['airfoils_coord_x'][:,i]
            fst_vt['AeroDyn15']['af_coord'][i]['y']  = inputs['airfoils_coord_y'][:,i]
            fst_vt['AeroDyn15']['rthick'][i]         = inputs['rthick'][i]
                
        # AeroDyn spanwise output positions
        r = r/r[-1]
        r_out_target = [0.0, 0.1, 0.20, 0.40, 0.6, 0.75, 0.85, 0.925, 1.0]
        idx_out = [np.argmin(abs(r-ri)) for ri in r_out_target]
        R_out = [fst_vt['AeroDynBlade']['BlSpn'][i] for i in idx_out]
        
        fst_vt['AeroDyn15']['BlOutNd'] = [str(idx+1) for idx in idx_out]
        fst_vt['AeroDyn15']['NBlOuts'] = len(idx_out)

        return fst_vt, R_out

    def DLC_creation(self, inputs, discrete_inputs, fst_vt):
        # Case Generations

        TMax = 99999. # Overwrite runtime if TMax is less than predefined DLC length (primarily for debugging purposes)
        # TMax = 5.

        list_cases        = []
        list_casenames    = []
        required_channels = []
        case_keys         = []

        turbulence_class = discrete_inputs['turbulence_class']
        turbine_class    = discrete_inputs['turbine_class']

        if self.DLC_powercurve != None:
            self.U_init     = copy.deepcopy(inputs['U_init'])
            self.Omega_init = copy.deepcopy(inputs['Omega_init'])
            self.pitch_init = copy.deepcopy(inputs['pitch_init'])
            # self.max_omega  = min([inputs['control_maxTS'] / inputs['Rtip'], inputs['control_maxOmega']*np.pi/30.])*30/np.pi
            # print('U_init    ', self.U_init    )
            # print('Omega_init', self.Omega_init)
            # print('pitch_init', self.pitch_init)
            # for i, (Ui, Omegai, pitchi) in enumerate(zip(self.U_init, self.Omega_init, self.pitch_init)):
            #     if pitchi > 0. and Omegai < self.max_omega*0.99:
            #         self.pitch_init[i] = 0.
            # print('U_init    ', self.U_init    )
            # print('Omega_init', self.Omega_init)
            # print('pitch_init', self.pitch_init)

            list_cases_PwrCrv, list_casenames_PwrCrv, requited_channels_PwrCrv = self.DLC_powercurve(fst_vt, self.FAST_runDirectory, self.FAST_namingOut, TMax, turbine_class, turbulence_class, inputs['Vrated'], U_init=self.U_init, Omega_init=self.Omega_init, pitch_init=self.pitch_init, V_R25=inputs['V_R25'])
            list_cases        += list_cases_PwrCrv
            list_casenames    += list_casenames_PwrCrv
            required_channels += requited_channels_PwrCrv
            case_keys         += [1]*len(list_cases_PwrCrv)    

        if self.DLC_gust != None:
            list_cases_gust, list_casenames_gust, requited_channels_gust = self.DLC_gust(fst_vt, self.FAST_runDirectory, self.FAST_namingOut, TMax, turbine_class, turbulence_class, inputs['V_mean_iec'], U_init=inputs['U_init'], Omega_init=inputs['Omega_init'], pitch_init=inputs['pitch_init'])
            list_cases        += list_cases_gust
            list_casenames    += list_casenames_gust
            required_channels += requited_channels_gust
            case_keys         += [2]*len(list_cases_gust)

        if self.DLC_extrm != None:
            list_cases_rated, list_casenames_rated, requited_channels_rated = self.DLC_extrm(fst_vt, self.FAST_runDirectory, self.FAST_namingOut, TMax, turbine_class, turbulence_class, inputs['Vextreme'])
            list_cases        += list_cases_rated
            list_casenames    += list_casenames_rated
            required_channels += requited_channels_rated
            case_keys         += [3]*len(list_cases_rated)

        if self.DLC_turbulent != None:
            if self.mpi_run:
                list_cases_turb, list_casenames_turb, requited_channels_turb = self.DLC_turbulent(fst_vt, self.FAST_runDirectory, self.FAST_namingOut, TMax, turbine_class, turbulence_class, inputs['Vrated'], U_init=inputs['U_init'], Omega_init=inputs['Omega_init'], pitch_init=inputs['pitch_init'], Turbsim_exe=self.Turbsim_exe, debug_level=self.debug_level, cores=self.cores, mpi_run=self.mpi_run, mpi_comm_map_down=self.mpi_comm_map_down)
            else:
                list_cases_turb, list_casenames_turb, requited_channels_turb = self.DLC_turbulent(fst_vt, self.FAST_runDirectory, self.FAST_namingOut, TMax, turbine_class, turbulence_class, inputs['Vrated'], U_init=inputs['U_init'], Omega_init=inputs['Omega_init'], pitch_init=inputs['pitch_init'], Turbsim_exe=self.Turbsim_exe, debug_level=self.debug_level, cores=self.cores)
            list_cases        += list_cases_turb
            list_casenames    += list_casenames_turb
            required_channels += requited_channels_turb
            case_keys         += [4]*len(list_cases_turb)

        required_channels = sorted(list(set(required_channels)))
        channels_out = {}
        for var in required_channels:
            channels_out[var] = True

        return list_cases, list_casenames, channels_out, case_keys


    def run_FAST(self, fst_vt, case_list, case_name_list, channels):

        # FAST wrapper setup
        fastBatch = runFAST_pywrapper_batch(FAST_ver=self.FAST_ver)

        fastBatch.FAST_exe          = self.FAST_exe
        fastBatch.FAST_runDirectory = self.FAST_runDirectory
        fastBatch.FAST_InputFile    = self.FAST_InputFile
        fastBatch.FAST_directory    = self.FAST_directory
        fastBatch.debug_level       = self.debug_level
        fastBatch.dev_branch        = self.dev_branch
        fastBatch.fst_vt            = fst_vt
        fastBatch.post              = return_timeseries

        fastBatch.case_list         = case_list
        fastBatch.case_name_list    = case_name_list
        fastBatch.channels          = channels

        # Run FAST
        if self.mpi_run:
            FAST_Output = fastBatch.run_mpi(self.mpi_comm_map_down)
        else:
            if self.cores == 1:
                FAST_Output = fastBatch.run_serial()
            else:
                FAST_Output = fastBatch.run_multi(self.cores)

        self.fst_vt = fst_vt

        return FAST_Output

    def write_FAST(self, fst_vt, discrete_outputs):
        writer                   = InputWriter_OpenFAST(FAST_ver=self.FAST_ver)
        writer.fst_vt            = fst_vt
        writer.FAST_runDirectory = self.FAST_runDirectory
        writer.FAST_namingOut    = self.FAST_namingOut
        writer.dev_branch        = self.dev_branch
        writer.execute()

        discrete_outputs['FASTpref_updated'] = copy.deepcopy(self.FASTpref)
        discrete_outputs['FASTpref_updated']['FAST_runDirectory'] = self.FAST_runDirectory
        discrete_outputs['FASTpref_updated']['FAST_directory']    = self.FAST_runDirectory
        discrete_outputs['FASTpref_updated']['FAST_InputFile']    = os.path.split(writer.FAST_InputFileOut)[-1]

        discrete_outputs['model_updated'] = True
        if self.debug_level > 0:
            print('RAN UPDATE: ', self.FAST_runDirectory, self.FAST_namingOut)

    def post_process(self, FAST_Output, case_keys, R_out, inputs, discrete_inputs, outputs):

        def post_gust(data, case_type):

            if case_type == 2:
                t_s = min(max(data['Time'][0], 30.), data['Time'][-2])
                t_e = min(data['Time'][-1], 90.)
                idx_s = list(data['Time']).index(t_s)
                idx_e = list(data['Time']).index(t_e)
            else:
                idx_s = 0
                idx_e = -1

            # Tip Deflections
            # return tip x,y,z for max out of plane deflection
            blade_max_tip = np.argmax([max(data['TipDxc1'][idx_s:idx_e]), max(data['TipDxc2'][idx_s:idx_e]), max(data['TipDxc3'][idx_s:idx_e])])
            if blade_max_tip == 0:
                tip_var = ["TipDxc1", "TipDyc1", "TipDzc1"]
            elif blade_max_tip == 1:
                tip_var = ["TipDxc2", "TipDyc2", "TipDzc2"]
            elif blade_max_tip == 2:
                tip_var = ["TipDxc3", "TipDyc3", "TipDzc3"]
            idx_max_tip = np.argmax(data[tip_var[0]][idx_s:idx_e])
            outputs['dx_defl'] = data[tip_var[0]][idx_s+idx_max_tip]
            outputs['dy_defl'] = data[tip_var[1]][idx_s+idx_max_tip]
            outputs['dz_defl'] = data[tip_var[2]][idx_s+idx_max_tip]

            # Root bending moments
            # return root bending moment for blade with the highest blade bending moment magnitude
            root_bending_moment_1 = np.sqrt(data['RootMxc1'][idx_s:idx_e]**2. + data['RootMyc1'][idx_s:idx_e]**2. + data['RootMzc1'][idx_s:idx_e]**2.)
            root_bending_moment_2 = np.sqrt(data['RootMxc2'][idx_s:idx_e]**2. + data['RootMyc2'][idx_s:idx_e]**2. + data['RootMzc2'][idx_s:idx_e]**2.)
            root_bending_moment_3 = np.sqrt(data['RootMxc3'][idx_s:idx_e]**2. + data['RootMyc3'][idx_s:idx_e]**2. + data['RootMzc3'][idx_s:idx_e]**2.)
            root_bending_moment_max       = [max(root_bending_moment_1), max(root_bending_moment_2), max(root_bending_moment_3)]
            root_bending_moment_idxmax    = [np.argmax(root_bending_moment_1), np.argmax(root_bending_moment_2), np.argmax(root_bending_moment_3)]
            blade_root_bending_moment_max = np.argmax(root_bending_moment_max)

            outputs['root_bending_moment'] = root_bending_moment_max[blade_root_bending_moment_max]*1.e3
            idx = root_bending_moment_idxmax[blade_root_bending_moment_max]
            if blade_root_bending_moment_max == 0:
                outputs['Mxyz'] = np.array([data['RootMxc1'][idx_s+idx]*1.e3, data['RootMyc1'][idx_s+idx]*1.e3, data['RootMzc1'][idx_s+idx]*1.e3])
            elif blade_root_bending_moment_max == 1:
                outputs['Mxyz'] = np.array([data['RootMxc2'][idx_s+idx]*1.e3, data['RootMyc2'][idx_s+idx]*1.e3, data['RootMzc2'][idx_s+idx]*1.e3])
            elif blade_root_bending_moment_max == 2:
                outputs['Mxyz'] = np.array([data['RootMxc3'][idx_s+idx]*1.e3, data['RootMyc3'][idx_s+idx]*1.e3, data['RootMzc3'][idx_s+idx]*1.e3])

        def post_extreme(data, case_type):

            if case_type == 3:
                t_s = min(max(data['Time'][0], 30.), data['Time'][-2])
                t_e = min(data['Time'][-1], 90.)
                idx_s = list(data['Time']).index(t_s)
                idx_e = list(data['Time']).index(t_e)
            else:
                idx_s = 0
                idx_e = -1

            Time = data['Time'][idx_s:idx_e]
            var_Fx = ["B1N1Fx", "B1N2Fx", "B1N3Fx", "B1N4Fx", "B1N5Fx", "B1N6Fx", "B1N7Fx", "B1N8Fx", "B1N9Fx"]
            var_Fy = ["B1N1Fy", "B1N2Fy", "B1N3Fy", "B1N4Fy", "B1N5Fy", "B1N6Fy", "B1N7Fy", "B1N8Fy", "B1N9Fy"]
            for i, (varFxi, varFyi) in enumerate(zip(var_Fx, var_Fy)):
                if i == 0:
                    Fx = np.array(data[varFxi][idx_s:idx_e])
                    Fy = np.array(data[varFyi][idx_s:idx_e])
                else:
                    Fx = np.column_stack((Fx, np.array(data[varFxi][idx_s:idx_e])))
                    Fy = np.column_stack((Fy, np.array(data[varFyi][idx_s:idx_e])))

            Fx_sum = np.zeros_like(Time)
            Fy_sum = np.zeros_like(Time)
            for i in range(len(Time)):
                Fx_sum[i] = np.trapz(Fx[i,:], R_out)
                Fy_sum[i] = np.trapz(Fy[i,:], R_out)
            idx_max_strain = np.argmax(np.sqrt(Fx_sum**2.+Fy_sum**2.))

            Fx = [data[Fxi][idx_max_strain] for Fxi in var_Fx]
            Fy = [data[Fyi][idx_max_strain] for Fyi in var_Fy]
            spline_Fx = PchipInterpolator(R_out, Fx)
            spline_Fy = PchipInterpolator(R_out, Fy)

            r = inputs['r']-inputs['Rhub']
            Fx_out = spline_Fx(r)
            Fy_out = spline_Fy(r)
            Fz_out = np.zeros_like(Fx_out)

            outputs['loads_Px'] = Fx_out
            outputs['loads_Py'] = Fy_out*-1.
            outputs['loads_Pz'] = Fz_out

            outputs['loads_Omega'] = data['RotSpeed'][idx_max_strain]
            outputs['loads_pitch'] = data['BldPitch1'][idx_max_strain]
            outputs['loads_azimuth'] = data['Azimuth'][idx_max_strain]

        # def post_AEP_fit(data):
        #     def my_cubic(f, x):
        #         return np.array([f[3]+ f[2]*xi + f[1]*xi**2. + f[0]*xi**3. for xi in x])

        #     U = np.array([np.mean(datai['Wind1VelX']) for datai in data])
        #     P = np.array([np.mean(datai['GenPwr']) for datai in data])*1000.
        #     P_coef = np.polyfit(U, P, 3)

        #     P_out = my_cubic(P_coef, inputs['V_out'])
        #     np.place(P_out, P_out>inputs['control_ratedPower'], inputs['control_ratedPower'])
        #     outputs['P_out'] = P_out

        #     # import matplotlib.pyplot as plt
        #     # plt.plot(U, P, 'o')
        #     # plt.plot(inputs['V_out'], outputs['P_out'])            
        #     # plt.show()

        def post_AEP(data):
            U = list(sorted([4., 6., 8., 9., 10., 10.5, 11., 11.5, 11.75, 12., 12.5, 13., 14., 19., 25., inputs['Vrated']]))
            if inputs['V_R25'] != 0.:
                U.append(inputs['V_R25'])
                U = list(sorted(U))
            U = np.array(U)

            U_below = [Vi for Vi in U if Vi <= inputs['Vrated']]
            # P_below = np.array([np.mean(datai['GenPwr'])*1000. for datai in data])
            P_below = np.array([np.mean(datai['GenPwr'])*1000. for datai, Vi in zip(data, U) if Vi <= inputs['Vrated']])
            np.place(P_below, P_below>inputs['control_ratedPower'], inputs['control_ratedPower'])

            U_rated = [Vi for Vi in U if Vi > inputs['Vrated']]
            P_rated = [inputs['control_ratedPower']]*len(U_rated)

            if len(U_below) < len(U):
                P_fast = np.array(P_below.tolist() + P_rated)
            else:
                P_fast = P_below

            data_rated = data[-1]

            # U_fit = np.array([4.,8.,9.,10.])

            ## Find rated 
            # def my_cubic(f, x):
                # return np.array([f[3]+ f[2]*xi + f[1]*xi**2. + f[0]*xi**3. for xi in x])

            # idx_fit = [U.tolist().index(Ui) for Ui in U_fit]
            # P_fit = np.array([np.mean(data[i]['GenPwr']) for i in idx_fit])
            # P_coef = np.polyfit(U_fit, P_fit, 3)

            # P_find_rated = my_cubic(P_coef, inputs['V_out'])
            # np.place(P_find_rated, P_find_rated>inputs['control_ratedPower'], inputs['control_ratedPower'])
            # idx_rated = min([i for i, Pi in enumerate(P_find_rated) if Pi*1000 >= inputs['control_ratedPower']])
            # outputs['rated_V'] = inputs['V_out'][idx_rated]

            # if outputs['rated_V'] not in U:
            #     ## Run Rated
            #     TMax = 99999.
            #     # TMax = 10.
            #     turbulence_class = TURBULENCE_CLASS[inputs['turbulence_class']]
            #     turbine_class    = TURBINE_CLASS[inputs['turbine_class']]
            #     list_cases_rated, list_casenames_rated, requited_channels_rated = RotorSE_rated(self.fst_vt, self.FAST_runDirectory, self.FAST_namingOut, TMax, turbine_class, turbulence_class, outputs['rated_V'], U_init=self.U_init, Omega_init=self.Omega_init, pitch_init=np.zeros_like(self.Omega_init))
            #     requited_channels_rated = sorted(list(set(requited_channels_rated)))
            #     channels_out = {}
            #     for var in requited_channels_rated:
            #         channels_out[var] = True
            #     data_rated = self.run_FAST(self.fst_vt, list_cases_rated, list_casenames_rated, channels_out)[0]

            #     ## Sort in Rated Power
            #     U_wR = []
            #     data_wR = []
            #     U_added = False
            #     for i in range(len(U)):
            #         if outputs['rated_V']<U[i] and U_added == False:
            #             U_wR.append(outputs['rated_V'])
            #             data_wR.append(data_rated)
            #             U_added = True
            #         U_wR.append(U[i])
            #         data_wR.append(data[i])
            # else:
            #     U_wR = U

            # P_fast = np.array([np.mean(datai['GenPwr']) for datai in data_wR])*1000.
            # for i, (Pi, Vi) in enumerate(zip(P_fast, U_wR)):
            #     if Vi > outputs['rated_V']:
            #         if np.abs((Pi-inputs['control_ratedPower'])/inputs['control_ratedPower']) > 0.2:
            #             P_fast[i] = inputs['control_ratedPower']
            #             above_rate_power_warning = "FAST instability expected at U=%f m/s, abs(outputted power) > +/-20%% of rated power.  Replaceing %f with %f"%(Vi, Pi, inputs['control_ratedPower'])
            #             warnings.warn(above_rate_power_warning)

            # P_spline = PchipInterpolator(U_wR, P_fast)

            P_spline = PchipInterpolator(U, P_fast)

            P_out = P_spline(inputs['V_out'])
            # np.place(P_out, P_out>inputs['control_ratedPower'], inputs['control_ratedPower'])
            outputs['P_out'] = P_out

            P = P_spline(inputs['V'])
            # np.place(P, P>inputs['control_ratedPower'], inputs['control_ratedPower'])
            outputs['P'] = P


            outputs['Cp']          = np.mean(data_rated["RtAeroCp"])
            outputs['rated_V']     = np.mean(data_rated["Wind1VelX"])
            outputs['rated_Omega'] = np.mean(data_rated["RotSpeed"])
            outputs['rated_pitch'] = np.mean(data_rated["BldPitch1"])
            outputs['rated_T']     = np.mean(data_rated["RotThrust"])*1000
            outputs['rated_Q']     = np.mean(data_rated["RotTorq"])*1000

            # import matplotlib.pyplot as plt
            # plt.plot(U, P, 'o')
            # plt.plot(inputs['V_out'], outputs['P_out'])            
            # plt.show()

        ############

        Gust_Outputs = False
        Extreme_Outputs = False
        AEP_Outputs = False
        #
        for casei in case_keys:
            if Gust_Outputs and Extreme_Outputs:
                break

            if casei == 1:
                # power curve
                if AEP_Outputs:
                    pass
                else:
                    idx_AEP = [i for i, casej in enumerate(case_keys) if casej==1]
                    data = [datai for i, datai in enumerate(FAST_Output) if i in idx_AEP]
                    post_AEP(data)
                    AEP_Outputs = True

            if casei in [2]:
                # gust: return tip deflections and bending moments
                idx_gust = case_keys.index(casei)
                data = FAST_Output[idx_gust]
                post_gust(data, casei)
                Gust_Outputs = True

            if casei in [3]:
                # extreme wind speed: return aeroloads for strains
                idx_extreme = case_keys.index(casei)
                data = FAST_Output[idx_extreme]
                post_extreme(data, casei)
                Extreme_Outputs = True

            if casei in [4]:
                # turbulent wind with multiplt seeds
                idx_turb = [i for i, casej in enumerate(case_keys) if casej==4]
                data_concat = {}
                for i, fast_out_idx in enumerate(idx_turb):
                    datai = FAST_Output[idx_turb[fast_out_idx]]

                    for var in datai.keys():
                        if i == 0:
                            data_concat[var] = []
                        data_concat[var].extend(list(datai[var]))

                for var in data_concat.keys():
                    data_concat[var] = np.array(data_concat[var])

                post_gust(data_concat, casei)
                post_extreme(data_concat, casei)
                Gust_Outputs = True
                Extreme_Outputs = True





                




import unittest

from wisdem.test.test_commonse import test_WindWaveDrag
from wisdem.test.test_commonse import test_enum
from wisdem.test.test_commonse import test_environment
from wisdem.test.test_commonse import test_frustum
from wisdem.test.test_commonse import test_tube
from wisdem.test.test_commonse import test_utilities
from wisdem.test.test_commonse import test_utilizationSupplement
from wisdem.test.test_commonse import test_vertical_cylinder

import numpy as np
import numpy.testing as npt

def suite():
    suite = unittest.TestSuite( (test_WindWaveDrag.suite(),
                                 test_enum.suite(),
                                 test_environment.suite(),
                                 test_frustum.suite(),
                                 test_tube.suite(),
                                 test_utilities.suite(),
                                 test_utilizationSupplement.suite(),
                                 test_vertical_cylinder.suite()
    ) )
    return suite


if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())
        
